#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2019 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""Reader for the old NWCSAF/Geo (v2013 and earlier) cloud product format.

References:
   - The NWCSAF GEO 2013 products documentation:
     http://www.nwcsaf.org/web/guest/archive - Search for Code "ICD/3"; Type
     "MSG" and the box to the right should say 'Status' (which means any
     status). Version 7.0 seems to be for v2013

     http://www.nwcsaf.org/aemetRest/downloadAttachment/2623

"""

import datetime as dt
import logging

import h5py
import numpy as np
from pyresample.geometry import AreaDefinition

from satpy.readers.core.hdf5 import HDF5FileHandler

logger = logging.getLogger(__name__)

PLATFORM_NAMES = {"MSG1": "Meteosat-8",
                  "MSG2": "Meteosat-9",
                  "MSG3": "Meteosat-10",
                  "MSG4": "Meteosat-11", }


class Hdf5NWCSAF(HDF5FileHandler):
    """NWCSAF MSG hdf5 reader."""

    def __init__(self, filename, filename_info, filetype_info):
        """Init method."""
        super(Hdf5NWCSAF, self).__init__(filename, filename_info,
                                         filetype_info)

        self.cache = {}

    def get_dataset(self, dataset_id, ds_info):
        """Load a dataset."""
        file_key = ds_info.get("file_key", dataset_id["name"])
        data = self[file_key]

        nodata = None
        if "SCALING_FACTOR" in data.attrs and "OFFSET" in data.attrs:
            dtype = np.dtype(data.data)
            if dataset_id["name"] in ["ctth_alti"]:
                data.attrs["valid_range"] = (0, 27000)
                data.attrs["_FillValue"] = np.nan

            if dataset_id["name"] in ["ctth_alti", "ctth_pres", "ctth_tempe", "ctth_effective_cloudiness"]:
                dtype = np.dtype("float32")
                nodata = 255

            if dataset_id["name"] in ["ct"]:
                data.attrs["valid_range"] = (0, 20)
                data.attrs["_FillValue"] = 255
                # data.attrs['palette_meanings'] = list(range(21))

            attrs = data.attrs
            scaled_data = (data * data.attrs["SCALING_FACTOR"] + data.attrs["OFFSET"]).astype(dtype)
            if nodata:
                scaled_data = scaled_data.where(data != nodata)
                scaled_data = scaled_data.where(scaled_data >= 0)
            data = scaled_data
            data.attrs = attrs

        for key in list(data.attrs.keys()):
            val = data.attrs[key]
            if isinstance(val, h5py.h5r.Reference):
                del data.attrs[key]

        return data

    def get_area_def(self, dsid):
        """Get the area definition of the datasets in the file."""
        if dsid["name"].endswith("_pal"):
            raise NotImplementedError

        cfac = self.file_content["/attr/CFAC"]
        lfac = self.file_content["/attr/LFAC"]
        coff = self.file_content["/attr/COFF"]
        loff = self.file_content["/attr/LOFF"]
        numcols = int(self.file_content["/attr/NC"])
        numlines = int(self.file_content["/attr/NL"])

        aex = get_area_extent(cfac, lfac, coff, loff, numcols, numlines)
        pname = self.file_content["/attr/PROJECTION_NAME"]
        proj = {}
        if pname.startswith("GEOS"):
            proj["proj"] = "geos"
            proj["a"] = "6378169.0"
            proj["b"] = "6356583.8"
            proj["h"] = "35785831.0"
            proj["lon_0"] = str(float(pname.split("<")[1][:-1]))
        else:
            raise NotImplementedError("Only geos projection supported yet.")

        area_def = AreaDefinition(self.file_content["/attr/REGION_NAME"],
                                  self.file_content["/attr/REGION_NAME"],
                                  pname,
                                  proj,
                                  numcols,
                                  numlines,
                                  aex)

        return area_def

    @property
    def start_time(self):
        """Return the start time of the object."""
        return dt.datetime.strptime(self.file_content["/attr/IMAGE_ACQUISITION_TIME"], "%Y%m%d%H%M")


def get_area_extent(cfac, lfac, coff, loff, numcols, numlines):
    """Get the area extent from msg parameters."""
    xur = (numcols - coff) * 2 ** 16 / (cfac * 1.0)
    xur = np.deg2rad(xur) * 35785831.0
    xll = (-1 - coff) * 2 ** 16 / (cfac * 1.0)
    xll = np.deg2rad(xll) * 35785831.0
    xres = (xur - xll) / numcols
    xur, xll = xur - xres / 2, xll + xres / 2
    yll = (numlines - loff) * 2 ** 16 / (-lfac * 1.0)
    yll = np.deg2rad(yll) * 35785831.0
    yur = (-1 - loff) * 2 ** 16 / (-lfac * 1.0)
    yur = np.deg2rad(yur) * 35785831.0
    yres = (yur - yll) / numlines
    yll, yur = yll + yres / 2, yur - yres / 2

    return xll, yll, xur, yur



# Copyright (c) 2025 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Utilities for writers."""
from __future__ import annotations

import os
import warnings
from collections.abc import Iterable
from typing import Any

import dask
import numpy as np
from dask import array as da
from dask.delayed import Delayed


def split_results(
        results: Iterable[list[da.Array | Delayed] | tuple[list[da.Array], list[Any]]],
) -> tuple[list[da.Array], list[Any], list[da.Array | Delayed]]:
    """Split results.

    Get sources, targets, and objects to be computed into separate lists from a list of
    results collected from one or more writers. This function will treat dask
    Arrays and Delayed objects with no targets as dask collections to be
    computed. Otherwise dask Arrays can be supplied with an associated target
    array-like file object and will be passed to :func:`dask.array.store`.

    We assume that the provided input is a list containing any combination of:

    1. List of dask Arrays (to be computed, not stored).
    2. List of Delayed objects.
    3. A two-element tuple where the first element is a collection of dask
       Arrays and the second is a collection of array-like file objects of
       the same size.
    4. A list made up of the above.

    """
    sources = []
    targets = []
    delayeds_or_arrays: list[da.Array | Delayed] = []

    for result in results:
        if isinstance(result, tuple) and len(result) == 2 and isinstance(result[0], list):
            sources.extend(result[0])
            targets.extend(result[1])
        elif isinstance(result, list):
            delayeds_or_arrays.extend(result)
        else:
            # Satpy 1.0 error message:
            # raise ValueError(f"Unexpected result from Satpy writer: {result!r}")
            # Satpy <1.0 warning:
            warnings.warn(f"Unexpected result from Satpy writer: {result!r}. "
                          "This form of result is no longer supported and will be removed in Satpy 1.0. "
                          "See docstring for expected values and https://github.com/pytroll/satpy/pull/3215 "
                          "for details.",
                          stacklevel=2)
            flat_results = _flatten(result)
            sources.extend([res for res in flat_results if isinstance(res, da.Array)])
            targets.extend([res for res in flat_results if not isinstance(res, (da.Array, Delayed))])
            delayeds_or_arrays.extend([res for res in flat_results if isinstance(res, Delayed)])
    return sources, targets, delayeds_or_arrays


def _flatten(results):
    # Remove in Satpy 1.0
    out = []
    if isinstance(results, (list, tuple)):
        for itm in results:
            out.extend(_flatten(itm))
        return out
    return [results]


def group_results_by_output_file(sources, targets):
    """Group results by output file.

    For writers that return sources and targets for ``compute=False``, split
    the results by output file.

    When not only the data but also GeoTIFF tags are dask arrays, then
    ``save_datasets(..., compute=False)``` returns a tuple of flat lists,
    where the second list consists of a mixture of ``RIOTag`` and ``RIODataset``
    objects (from trollimage).  In some cases, we may want to get a seperate
    delayed object for each file; for example, if we want to add a wrapper to do
    something with the file as soon as it's finished.  This function unflattens
    the flat lists into a list of (src, target) tuples.

    For example, to close files as soon as computation is completed::

        >>> @dask.delayed
        >>> def closer(obj, targs):
        ...     for targ in targs:
        ...         targ.close()
        ...     return obj
        >>> (srcs, targs) = sc.save_datasets(writer="ninjogeotiff", compute=False, **ninjo_tags)
        >>> for (src, targ) in group_results_by_output_file(srcs, targs):
        ...     delayed_store = da.store(src, targ, compute=False)
        ...     wrapped_store = closer(delayed_store, targ)
        ...     wrapped.append(wrapped_store)
        >>> compute_writer_results(wrapped)

    In the wrapper you can do other useful tasks, such as writing a log message
    or moving files to a different directory.

    .. warning::

        Adding a callback may impact runtime and RAM.  The pattern or cause is
        unclear.  Tests with FCI data show that for resampling with high RAM
        use (from around 15 GB), runtime increases when a callback is added.
        Tests with ABI or low RAM consumption rather show a decrease in runtime.
        More information, see `these GitHub comments
        <https://github.com/pytroll/satpy/pull/2281#issuecomment-1324910253>`_
        Users who find out more are encouraged to contact the Satpy developers
        with clues.

    Args:
        sources: List of sources (typically dask.array) as returned by
            :meth:`satpy.scene.Scene.save_datasets`.
        targets: List of targets (should be ``RIODataset`` or ``RIOTag``) as
            returned by :meth:`satpy.scene.Scene.save_datasets`.

    Returns:
        List of ``Tuple(List[sources], List[targets])`` with a length equal to
        the number of output files planned to be written by
        :meth:`satpy.scene.Scene.save_datasets`.
    """
    ofs = {}
    for (src, targ) in zip(sources, targets):
        fn = targ.rfile.path
        if fn not in ofs:
            ofs[fn] = ([], [])
        ofs[fn][0].append(src)
        ofs[fn][1].append(targ)
    return list(ofs.values())


def compute_writer_results(
        results: Iterable[list[da.Array | Delayed] | tuple[list[da.Array], list[Any]]],
) -> list[np.ndarray | str | os.PathLike | None]:
    """Compute all the given dask graphs ``results`` so that the files are saved.

    Args:
        results: Iterable of dask collections resulting from calls to
            ``scn.save_datasets(..., compute=False)``. The iterable passed
            should always **contain** the results of the writer/save_datasets
            call, not be the results themselves. Each collection of results from
            a writer should be either a list of dask containers to be computed
            (ex. Array or Delayed) or a 2-element tuple where the first element
            is a list of dask Array objects and the second element is a list of
            target file-like objects supporting ``__setitem__`` syntax. In the
            case of the 2-element tuple, these lists will be passed to
            :func:`dask.array.store` to write the Arrays to the targets.

    Returns:
        A list of the computed results. These are typically string or Path
        filenames that were created or None if the ``store`` function
        (see above) was called. If the dask collection to be computed
        is an array then the computed array will be in the returned list.

    """
    computed_results: list[np.ndarray | str | os.PathLike | None] = []
    if not results:
        return computed_results

    sources, targets, delayeds_or_arrays = split_results(results)

    # one or more writers have targets that we need to close in the future
    if targets:
        delayeds_or_arrays.append(da.store(sources, targets, compute=False))

    if delayeds_or_arrays:
        # replace Delayed's graph optimization function with the Array function
        # since a Delayed object here is only from the writer but the rest of
        # the tasks are dask array operations we want to fully optimize all
        # array operations. At the time of writing Array optimizations seem to
        # include the optimizations done for Delayed objects alone.
        with dask.config.set(delayed_optimization=dask.config.get("array_optimize", da.optimize)):
            computed_results.extend(da.compute(delayeds_or_arrays))

    if targets:
        for target in targets:
            if hasattr(target, "close"):
                target.close()

    return computed_results



# Copyright (c) 2017-2025 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""Unit testing the convolution enhancements functions."""

import numpy as np

from .utils import create_ch1, run_and_check_enhancement


def test_three_d_effect():
    """Test the three_d_effect enhancement function."""
    from satpy.enhancements.convolution import three_d_effect

    ch1 = create_ch1()
    expected = np.array([[
        [np.nan, np.nan, -389.5, -294.5, 826.5],
        [np.nan, np.nan, 85.5, 180.5, 1301.5]]])
    run_and_check_enhancement(three_d_effect, ch1, expected)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2025 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""PACE OCI L1b reader.

This reader supports an optional argument to choose the "engine" for reading
the OCI netCDF4 files. By default, this reader uses the default xarray choice of
engine, as defined in the :func:`xarray.open_dataset` documentation`.

As an alternative, the user may wish to use the "h5netcdf" engine, but that is
not default as it typically prints many non-fatal but confusing error messages
to the terminal.
To choose between engines the user can  do as follows for the default::

    scn = Scene(filenames=my_files, reader="olci_l1b")

or as follows for the h5netcdf engine::

    scn = Scene(filenames=my_files,
                reader="olci_l1b", reader_kwargs={"engine": "h5netcdf"})

References:
    - :func:`xarray.open_dataset`

"""

import logging
from datetime import datetime

import numpy as np
import xarray as xr

from satpy._compat import cached_property
from satpy.readers.core.file_handlers import BaseFileHandler
from satpy.readers.core.remote import open_file_or_filename
from satpy.utils import get_legacy_chunk_size

CHUNK_SIZE = get_legacy_chunk_size()

logger = logging.getLogger(__name__)


class NCOCIL1B(BaseFileHandler):
    """The OCI reader base."""

    rows_name = "scans"
    cols_name = "pixels"

    @staticmethod
    def _sort_wvls(inwvls, bw):
        """Compute the bandwidth for each hyperspectral channel."""
        wvls = np.dstack((inwvls, inwvls, inwvls)).squeeze()

        # If we have a float, use that as the bandwidth for all bands.
        if type(bw) is float:
            wvls[:, 0] = wvls[:, 0] - bw
            wvls[:, 2] = wvls[:, 2] + bw
        # But otherwise, use the array as per-band bandwidths.
        else:
            wvls[:, 0] = wvls[:, 0] - bw / 2.
            wvls[:, 2] = wvls[:, 2] + bw / 2.

        return wvls / 1000.

    def __init__(self, filename, filename_info, filetype_info, **kwargs):
        """Init the oci reader base."""
        super().__init__(filename, filename_info, filetype_info)
        self._engine = kwargs.get("engine", None)
        self.sensor = "oci"

        # Get the per-band solar irradiance and central wavelength values, these are split across
        # three variables, one each for the blue, red and SWIR bands.
        self.irradiance = {"blue": self.nc["sensor_band_parameters"]["blue_solar_irradiance"].values,
                           "red": self.nc["sensor_band_parameters"]["red_solar_irradiance"].values,
                           "swir": self.nc["sensor_band_parameters"]["SWIR_solar_irradiance"].values}

        swir_bw = self.nc["sensor_band_parameters"]["SWIR_bandpass"].values

        self.wvls = {"blue": self._sort_wvls(self.nc["sensor_band_parameters"]["blue_wavelength"].values, 0.5),
                     "red": self._sort_wvls(self.nc["sensor_band_parameters"]["red_wavelength"].values, 0.5),
                     "swir": self._sort_wvls(self.nc["sensor_band_parameters"]["SWIR_wavelength"].values, swir_bw)}

        # Spatial resolution is hardcoded, same for all datasets.
        self.resolution = 1200

    @cached_property
    def nc(self):
        """Open the nc xr dataset."""
        f_obj = open_file_or_filename(self.filename)
        dataset = xr.open_datatree(f_obj,
                                   decode_cf=True,
                                   mask_and_scale=True,
                                   engine=self._engine,
                                   chunks={self.cols_name: CHUNK_SIZE,
                                           self.rows_name: CHUNK_SIZE})
        return dataset

    def available_datasets(self, configured_datasets=None):
        """Form the names for the available datasets."""
        for bnd in ["blue", "red", "SWIR"]:
            shper = self.nc["observation_data"][f"rhot_{bnd}"].shape
            for i in range(0, shper[0]):
                yield True, self._retr_dsinfo_chans(bnd, i, {"cal": "refl", "dst": "chan", "vatype": "rhot"})
                yield True, self._retr_dsinfo_chans(bnd, i, {"cal": "radi", "dst": "chan", "vatype": "rhot"})
                yield True, self._retr_dsinfo_chans(bnd, i, {"cal": "", "dst": "qual", "vatype": "qual"})

        for is_avail, ds_info in (configured_datasets or []):
            yield True, ds_info

    def _retr_dsinfo_chans(self, band, i, idic):
        """Retrieve the ds info for a given channel."""
        ds_info = {"file_type": self.filetype_info["file_type"],
                   "resolution": self.resolution,
                   "name": f"{idic['dst']}_{band.lower()}_{self.wvls[band.lower()][i][1] * 1000:4.0f}".replace(" ", ""),
                   "wavelength": [self.wvls[band.lower()][i][0],
                                  self.wvls[band.lower()][i][1],
                                  self.wvls[band.lower()][i][2]],
                   "file_key": f"{idic['vatype']}_{band}",
                   "ds_key": band,
                   "idx": i,
                   "grp_key": "observation_data",
                   "coordinates": ("longitude", "latitude")}

        if idic["cal"] == "refl":
            ds_info["standard_name"] = "toa_bidirectional_reflectance"
            ds_info["units"] = "%"
            ds_info["calibration"] = "reflectance"
        elif idic["cal"] == "radi":
            ds_info["standard_name"] = "toa_outgoing_radiance_per_unit_wavelength"
            ds_info["units"] = "W m-2 sr-1 um-1"
            ds_info["calibration"] = "radiance"
        elif idic["dst"] == "qual":
            ds_info["standard_name"] = "quality_flags"
            ds_info["units"] = "1"
        return ds_info

    def _sort_cal(self, variable, caltype, info):
        """Apply calibration to the data, if needed, and update attrs."""
        if caltype == "reflectance":
            variable = variable * 100.
            variable.attrs["units"] = "%"
            variable.attrs["standard_name"] = "toa_bidirectional_reflectance"
        elif caltype == "radiance":
            sza = self.nc["geolocation_data/solar_zenith"]
            esd = self.nc.attrs["earth_sun_distance_correction"]
            irr = self.irradiance[info["ds_key"].lower()][info["idx"]]
            variable = (variable * irr * np.cos(np.radians(sza))) / (esd * np.pi)
            variable.attrs["units"] = "W m-2 sr-1 um-1"
            variable.attrs["standard_name"] = "toa_outgoing_radiance_per_unit_wavelength"

        return variable

    def get_dataset(self, key, info):
        """Load a dataset."""
        logger.debug("Reading %s.", key["name"])
        if "file_key" not in info:
            variable = self.nc[info["grp_key"]][key["name"]]
        else:
            if "idx" in info:
                variable = self.nc[info["grp_key"]][info["file_key"]][info["idx"]]
            else:
                variable = self.nc[info["grp_key"]][info["file_key"]]

        # Datatree messes with the coordinates, for lats + lons we need to remove the coords.
        if key["name"] in ["longitude", "latitude"]:
            variable = variable.reset_coords(names="longitude", drop=True)
            variable = variable.reset_coords(names="latitude", drop=True)
        variable.attrs["standard_name"] = info["standard_name"]
        if "calibration" in key:
            variable = self._sort_cal(variable, key["calibration"], info)

        return variable.rename({self.cols_name: "x", self.rows_name: "y"})

    @property
    def start_time(self):
        """Start time property."""
        st_dt = self.nc.attrs["time_coverage_start"]
        return datetime.strptime(st_dt, "%Y-%m-%dT%H:%M:%S.%fZ")

    @property
    def end_time(self):
        """End time property."""
        """Start time property."""
        st_dt = self.nc.attrs["time_coverage_end"]
        return datetime.strptime(st_dt, "%Y-%m-%dT%H:%M:%S.%fZ")



# Copyright (c) 2023 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""A reader for OSI-SAF level 3 products in netCDF format."""

import datetime as dt
import logging

from satpy.readers.core.netcdf import NetCDF4FileHandler

logger = logging.getLogger(__name__)


class OSISAFL3NCFileHandler(NetCDF4FileHandler):
    """Reader for the OSISAF l3 netCDF format."""
    def _get_ease_grid(self):
        """Set up the EASE grid."""
        from pyresample import create_area_def

        proj4str = self["Lambert_Azimuthal_Grid/attr/proj4_string"]
        x_size = self["/dimension/xc"]
        y_size = self["/dimension/yc"]
        p_lowerleft_lat = self["lat"].values[y_size - 1, 0]
        p_lowerleft_lon = self["lon"].values[y_size - 1, 0]
        p_upperright_lat = self["lat"].values[0, x_size - 1]
        p_upperright_lon = self["lon"].values[0, x_size - 1]
        area_extent = [p_lowerleft_lon, p_lowerleft_lat, p_upperright_lon, p_upperright_lat]
        area_def = create_area_def(area_id="osisaf_lambert_azimuthal_equal_area",
                                   description="osisaf_lambert_azimuthal_equal_area",
                                   proj_id="osisaf_lambert_azimuthal_equal_area",
                                   projection=proj4str, width=x_size, height=y_size, area_extent=area_extent,
                                   units="deg")
        return area_def

    def _get_geographic_grid(self):
        """Set up the EASE grid."""
        from pyresample import create_area_def

        x_size = self["/dimension/lon"]
        y_size = self["/dimension/lat"]
        lat_0 = self["lat"].min()
        lon_0 = self["lon"].min()
        lat_1 = self["lat"].max()
        lon_1 = self["lon"].max()
        area_extent = [lon_0, lat_1, lon_1, lat_0]
        area_def = create_area_def(area_id="osisaf_geographic_area",
                                   description="osisaf_geographic_area",
                                   proj_id="osisaf_geographic_area",
                                   projection="+proj=lonlat", width=x_size, height=y_size, area_extent=area_extent,
                                   units="deg")
        return area_def

    def _get_polar_stereographic_grid(self):
        """Set up the polar stereographic grid."""
        from pyresample import create_area_def
        try:
            proj4str = self["Polar_Stereographic_Grid/attr/proj4_string"]
        except KeyError:
            # Some products don't have the proj str, so we construct it ourselves
            sma = self["Polar_Stereographic_Grid/attr/semi_major_axis"]
            smb = self["Polar_Stereographic_Grid/attr/semi_minor_axis"]
            lon_0 = self["Polar_Stereographic_Grid/attr/straight_vertical_longitude_from_pole"]
            lat_0 = self["Polar_Stereographic_Grid/attr/latitude_of_projection_origin"]
            lat_ts = self["Polar_Stereographic_Grid/attr/standard_parallel"]
            proj4str = f"+a={sma} +b={smb} +lat_ts={lat_ts} +lon_0={lon_0} +proj=stere +lat_0={lat_0}"
        x_size = self["/dimension/xc"]
        y_size = self["/dimension/yc"]
        p_lowerleft_lat = self["lat"].values[y_size - 1, 0]
        p_lowerleft_lon = self["lon"].values[y_size - 1, 0]
        p_upperright_lat = self["lat"].values[0, x_size - 1]
        p_upperright_lon = self["lon"].values[0, x_size - 1]
        area_extent = [p_lowerleft_lon, p_lowerleft_lat, p_upperright_lon, p_upperright_lat]
        area_def = create_area_def(area_id="osisaf_polar_stereographic",
                                   description="osisaf_polar_stereographic",
                                   proj_id="osisaf_polar_stereographic",
                                   projection=proj4str, width=x_size, height=y_size, area_extent=area_extent,
                                   units="deg")
        return area_def


    def _get_finfo_grid(self):
        """Get grid in case of filename info being used."""
        if self.filename_info["grid"] == "ease":
            self.area_def = self._get_ease_grid()
            return self.area_def
        elif self.filename_info["grid"] == "polstere" or self.filename_info["grid"] == "stere":
            self.area_def = self._get_polar_stereographic_grid()
            return self.area_def
        else:
            raise ValueError(f"Unknown grid type: {self.filename_info['grid']}")

    def _get_ftype_grid(self):
        """Get grid in case of filetype info being used."""
        if self.filetype_info["file_type"] == "osi_radflux_grid":
            self.area_def = self._get_geographic_grid()
            return self.area_def
        elif self.filetype_info["file_type"] in ["osi_sst", "osi_sea_ice_conc"]:
            self.area_def = self._get_polar_stereographic_grid()
            return self.area_def

    def get_area_def(self, area_id):
        """Get the area definition, which varies depending on file type and structure."""
        if "grid" in self.filename_info:
            return self._get_finfo_grid()
        else:
            return self._get_ftype_grid()


    def _get_ds_units(self, ds_info, var_path):
        """Find the units of the datasets."""
        file_units = ds_info.get("file_units")
        if file_units is None:
            file_units = self.get(var_path + "/attr/units")
            if file_units is None:
                file_units = 1
        return file_units

    def get_dataset(self, dataset_id, ds_info):
        """Load a dataset."""
        logger.debug(f"Reading {dataset_id['name']} from {self.filename}")
        var_path = ds_info.get("file_key", f"{dataset_id['name']}")

        shape = self[var_path + "/shape"]
        data = self[var_path]
        if shape[0] == 1:
            # Remove the time dimension from dataset
            data = data[0]

        file_units = self._get_ds_units(ds_info, var_path)

        # Try to get the valid limits for the data.
        # Not all datasets have these, so fall back on assuming no limits.
        valid_min = self.get(var_path + "/attr/valid_min")
        valid_max = self.get(var_path + "/attr/valid_max")
        if valid_min is not None and valid_max is not None:
            data = data.where(data >= valid_min)
            data = data.where(data <= valid_max)

        # Try to get the fill value for the data.
        # If there isn't one, assume all remaining pixels are valid.
        fill_value = self.get(var_path + "/attr/_FillValue")
        if fill_value is not None:
            data = data.where(data != fill_value)

        # Try to get the scale and offset for the data.
        # As above, not all datasets have these, so fall back on assuming no limits.
        scale_factor = self.get(var_path + "/attr/scale_factor")
        scale_offset = self.get(var_path + "/attr/add_offset")
        if scale_offset is not None and scale_factor is not None:
            data = (data * scale_factor + scale_offset)

        # Set proper dimension names
        if self.filetype_info["file_type"] == "osi_radflux_grid":
            data = data.rename({"lon": "x", "lat": "y"})
        else:
            data = data.rename({"xc": "x", "yc": "y"})

        ds_info.update({
            "units": ds_info.get("units", file_units),
            "platform_name": self._get_platname(),
            "sensor": self._get_instname()
        })
        ds_info.update(dataset_id.to_dict())
        data.attrs.update(ds_info)
        return data

    def _get_instname(self):
        """Get instrument name."""
        try:
            return self["/attr/instrument_name"]
        except KeyError:
            try:
                return self["/attr/sensor"]
            except KeyError:
                return "unknown_sensor"

    def _get_platname(self):
        """Get platform name."""
        try:
            return self["/attr/platform_name"]
        except KeyError:
            return self["/attr/platform"]

    @staticmethod
    def _parse_datetime(datestr):
        for dt_format in ("%Y-%m-%d %H:%M:%S","%Y%m%dT%H%M%SZ", "%Y-%m-%dT%H:%M:%SZ"):
            try:
                return dt.datetime.strptime(datestr, dt_format)
            except ValueError:
                continue
        raise ValueError(f"Unsupported date format: {datestr}")

    @property
    def start_time(self):
        """Get the start time."""
        poss_names = ["/attr/start_date", "/attr/start_time", "/attr/time_coverage_start"]
        for name in poss_names:
            start_t = self.get(name)
            if start_t is not None:
                break
        if start_t is None:
            raise ValueError("Unknown start time attribute.")
        return self._parse_datetime(start_t)

    @property
    def end_time(self):
        """Get the end time."""
        poss_names = ["/attr/stop_date", "/attr/stop_time", "/attr/time_coverage_end"]
        for name in poss_names:
            end_t = self.get(name)
            if end_t is not None:
                break
        if end_t is None:
            raise ValueError("Unknown stop time attribute.")
        return self._parse_datetime(end_t)



# Copyright (c) 2024 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Image comparison tests."""

import hdf5plugin  # noqa: F401  isort:skip
import os
import os.path
import warnings
from datetime import datetime
from glob import glob

import cv2
import dask
import numpy as np
from behave import given, then, when

import satpy
from satpy import Scene

ext_data_path = "/app/ext_data"
threshold = 2000

def before_all(context):
    """Define a before_all hook to create the timestamp and test results directory."""
    tm = datetime.now()
    context.timestamp = tm.strftime("%Y-%m-%d-%H-%M-%S")
    context.test_results_dir = f"{ext_data_path}/test_results/image_comparison/{context.timestamp}"
    os.makedirs(os.path.join(context.test_results_dir, "generated"), exist_ok=True)
    os.makedirs(os.path.join(context.test_results_dir, "difference"), exist_ok=True)

    # Write the timestamp to test_results.txt
    results_file = os.path.join(context.test_results_dir, "test_results.txt")
    with open(results_file, "a") as f:
        f.write(f"Test executed at {context.timestamp}.\n\n")

def setup_hooks():
    """Register the before_all hook."""
    from behave import use_fixture
    from behave.runner import Context

    use_fixture(before_all, Context)

setup_hooks()
@given("I have a {composite} reference image file from {satellite} resampled to {area}")
def step_given_reference_image(context, composite, satellite, area):
    """Prepare a reference image."""
    reference_image = f"satpy-reference-image-{satellite}-{composite}-{area}.png"
    context.reference_image = cv2.imread(f"{ext_data_path}/reference_images/{reference_image}")
    context.satellite = satellite
    context.composite = composite
    context.area = area


@when("I generate a new {composite} image file from {satellite} case {case} "
      "with {reader} for {area} resampling with {resampler} with clipping {clip}")
def step_when_generate_image(context, composite, satellite, case, reader, area, resampler, clip):
    """Generate test images."""
    os.environ["OMP_NUM_THREADS"] = os.environ["MKL_NUM_THREADS"] = "2"
    os.environ["PYTROLL_CHUNK_SIZE"] = "1024"
    warnings.simplefilter("ignore")
    dask.config.set(scheduler="threads", num_workers=4)

    # Get the list of satellite files to open
    filenames = glob(f"{ext_data_path}/satellite_data/{satellite}/{case}/*.nc")


    if "," in reader:
        reader = reader.split(",")
    with satpy.config.set({"readers.clip_negative_radiances": False if clip == "null" else clip}):
        scn = Scene(reader=reader, filenames=filenames)
        scn.load([composite])

    if area == "null":
        ls = scn
    else:
        ls = scn.resample(area, resampler=resampler)

    # Save the generated image in the generated folder
    generated_image_path = os.path.join(context.test_results_dir, "generated",
                                        f"generated_{context.satellite}_{context.composite}_{context.area}.png")
    ls.save_datasets(writer="simple_image", filename=generated_image_path)

    # Save the generated image in the context
    context.generated_image = cv2.imread(generated_image_path)


@then("the generated image should be the same as the reference image")
def step_then_compare_images(context):
    """Compare test image to reference image."""
    # Load the images
    imageA = cv2.cvtColor(context.reference_image, cv2.COLOR_BGR2GRAY)
    imageB = cv2.cvtColor(context.generated_image, cv2.COLOR_BGR2GRAY)
    # Ensure both images have the same dimensions
    if imageA.shape != imageB.shape:
        raise ValueError("Both images must have the same dimensions")
    array1 = np.array(imageA)
    array2 = np.array(imageB)
    # Perform pixel-wise comparison
    result_matrix = (array1 != array2).astype(np.uint8) * 255

    # Save the resulting numpy array as an image in the difference folder
    diff_image_path = os.path.join(context.test_results_dir, "difference",
                                   f"diff_{context.satellite}_{context.composite}.png")
    cv2.imwrite(diff_image_path, result_matrix)

    # Count non-zero pixels in the result matrix
    non_zero_count = np.count_nonzero(result_matrix)

    # Write the results to a file in the test results directory
    results_file = os.path.join(context.test_results_dir, "test_results.txt")
    with open(results_file, "a") as f:
        f.write(f"Test for {context.satellite} - {context.composite}\n")
        f.write(f"Non-zero pixel differences: {non_zero_count}\n")
        if non_zero_count < threshold:
            f.write(f"Result: Passed - {non_zero_count} pixel differences.\n\n")
        else:
            f.write(f"Result: Failed - {non_zero_count} pixel differences exceed the threshold of {threshold}.\n\n")

    # Assert that the number of differences is below the threshold
    assert non_zero_count < threshold, (f"Images are not similar enough. "
            f"{non_zero_count} pixel differences exceed the threshold of "
            f"{threshold}.")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2018 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Module for testing the satpy.readers.core.hdf5 module."""

import os
import unittest

import numpy as np

try:
    from satpy.readers.core.hdf5 import HDF5FileHandler
except ImportError:
    # fake the import so we can at least run the tests in this file
    HDF5FileHandler = object  # type: ignore


class FakeHDF5FileHandler(HDF5FileHandler):
    """Swap  HDF5 File Handler for reader tests to use."""

    def __init__(self, filename, filename_info, filetype_info, **kwargs):
        """Get fake file content from 'get_test_content'."""
        if HDF5FileHandler is object:
            raise ImportError("Base 'HDF5FileHandler' could not be "
                              "imported.")
        filename = str(filename)
        super(HDF5FileHandler, self).__init__(filename, filename_info, filetype_info)
        self.file_content = self.get_test_content(filename, filename_info, filetype_info)
        self.file_content.update(kwargs)

    def get_test_content(self, filename, filename_info, filetype_info):
        """Mimic reader input file content.

        Args:
            filename (str): input filename
            filename_info (dict): Dict of metadata pulled from filename
            filetype_info (dict): Dict of metadata from the reader's yaml config for this file type

        Returns: dict of file content with keys like:

            - 'dataset'
            - '/attr/global_attr'
            - 'dataset/attr/global_attr'
            - 'dataset/shape'

        """
        raise NotImplementedError("Fake File Handler subclass must implement 'get_test_content'")


class TestHDF5FileHandler(unittest.TestCase):
    """Test HDF5 File Handler Utility class."""

    def setUp(self):
        """Create a test HDF5 file."""
        import h5py
        h = h5py.File("test.h5", "w")
        # Create Group
        g1 = h.create_group("test_group")

        # Add datasets
        ds1_f = g1.create_dataset("ds1_f",
                                  shape=(10, 100),
                                  dtype=np.float32,
                                  data=np.arange(10. * 100).reshape((10, 100)))
        ds1_i = g1.create_dataset("ds1_i",
                                  shape=(10, 100),
                                  dtype=np.int32,
                                  data=np.arange(10 * 100).reshape((10, 100)))
        ds2_f = h.create_dataset("ds2_f",
                                 shape=(10, 100),
                                 dtype=np.float32,
                                 data=np.arange(10. * 100).reshape((10, 100)))
        ds2_i = h.create_dataset("ds2_i",
                                 shape=(10, 100),
                                 dtype=np.int32,
                                 data=np.arange(10 * 100).reshape((10, 100)))

        # Add attributes
        # shows up as a scalar array of bytes (shape=(), size=1)
        h.attrs["test_attr_str"] = "test_string"
        h.attrs["test_attr_byte"] = b"test_byte"
        h.attrs["test_attr_int"] = 0
        h.attrs["test_attr_float"] = 1.2
        # shows up as a numpy bytes object
        h.attrs["test_attr_str_arr"] = np.array(b"test_string2")
        g1.attrs["test_attr_str"] = "test_string"
        g1.attrs["test_attr_byte"] = b"test_byte"
        g1.attrs["test_attr_int"] = 0
        g1.attrs["test_attr_float"] = 1.2
        for d in [ds1_f, ds1_i, ds2_f, ds2_i]:
            d.attrs["test_attr_str"] = "test_string"
            d.attrs["test_attr_byte"] = b"test_byte"
            d.attrs["test_attr_int"] = 0
            d.attrs["test_attr_float"] = 1.2
            d.attrs["test_ref"] = d.ref
        self.var_attrs = list(d.attrs.keys())

        h.close()

    def tearDown(self):
        """Remove the previously created test file."""
        os.remove("test.h5")

    def test_all_basic(self):
        """Test everything about the HDF5 class."""
        import xarray as xr

        from satpy.readers.core.hdf5 import HDF5FileHandler
        file_handler = HDF5FileHandler("test.h5", {}, {})

        for ds_name in ("test_group/ds1_f", "test_group/ds1_i", "ds2_f", "ds2_i"):
            ds = file_handler[ds_name]
            attrs = ds.attrs
            assert ds.dtype == (np.float32 if ds_name.endswith("f") else np.int32)
            assert file_handler[ds_name + "/shape"] == (10, 100)
            assert attrs["test_attr_str"] == "test_string"
            assert attrs["test_attr_byte"] == "test_byte"
            assert attrs["test_attr_int"] == 0
            assert attrs["test_attr_float"] == 1.2
            assert file_handler[ds_name + "/attr/test_attr_str"] == "test_string"
            assert file_handler[ds_name + "/attr/test_attr_byte"] == "test_byte"
            assert file_handler[ds_name + "/attr/test_attr_int"] == 0
            assert file_handler[ds_name + "/attr/test_attr_float"] == 1.2

        assert file_handler["/attr/test_attr_str"] == "test_string"
        assert file_handler["/attr/test_attr_byte"] == "test_byte"
        assert file_handler["/attr/test_attr_str_arr"] == "test_string2"
        assert file_handler["/attr/test_attr_int"] == 0
        assert file_handler["/attr/test_attr_float"] == 1.2

        assert isinstance(file_handler.get("ds2_f"), xr.DataArray)
        assert file_handler.get("fake_ds") is None
        assert file_handler.get("fake_ds", "test") == "test"

        assert "ds2_f" in file_handler
        assert "fake_ds" not in file_handler

        assert isinstance(file_handler["ds2_f/attr/test_ref"], np.ndarray)

    def test_array_name_uniqueness(self):
        """Test the dask array generated from an hdf5 dataset stay constant and unique."""
        from satpy.readers.core.hdf5 import HDF5FileHandler
        file_handler = HDF5FileHandler("test.h5", {}, {})

        dsname = "test_group/ds1_f"

        assert file_handler[dsname].data.name == file_handler[dsname].data.name
        assert file_handler[dsname].data.name.startswith("/" + dsname)



# Copyright (c) 2023 - 2025 Pytroll Developers

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Reader for the level-1b data from the MWR sounder onboard AWS and EPS-STerna.

AWS = Arctic Weather Satellite. MWR = Microwave Radiometer.

AWS test data provided by ESA August 23, 2023.

Sample data for five orbits in September 2024 provided by ESA to the Science
Advisory Group for MWS and AWS, November 26, 2024.

Sample EPS-Sterna l1b format AWS data from 16 orbits the 9th of November 2024.

Continous feed (though restricted to the SAG members and selected European
users/evaluators) in the EUMETSAT Data Store of global AWS data from January
9th, 2025.

Example:
--------
Here is an example how to read the data in satpy:

.. code-block:: python

    from satpy import Scene
    from glob import glob

    filenames = glob("data/W_NO-KSAT-Tromso,SAT,AWS1-MWR-1B-RAD_C_OHB__*_G_O_20250110114708*.nc"
    scn = Scene(filenames=filenames, reader='aws1_mwr_l1b_nc')

    composites = ['mw183_humidity']
    dataset_names = composites + ['1']

    scn.load(dataset_names)
    print(scn['1'])
    scn.show('mw183_humidity')


As the file format for the EPS Sterna Level-1b is slightly different from the
ESA format, reading the EPS Sterna level-1b data uses a different reader, named
`eps_sterna_mwr_l1b_nc`. So, if specifying the reader name as in the above code
example, please provide the actual name for that data: eps_sterna_mwr_l1b_nc.


"""

import xarray as xr

from satpy.readers.core.netcdf import NetCDF4FileHandler

MWR_CHANNEL_NAMES = [str(i) for i in range(1, 20)]

NAVIGATION_DATASET_NAMES = ["satellite_zenith_horn1",
                            "satellite_zenith_horn2",
                            "satellite_zenith_horn3",
                            "satellite_zenith_horn4",
                            "solar_azimuth_horn1",
                            "solar_azimuth_horn2",
                            "solar_azimuth_horn3",
                            "solar_azimuth_horn4",
                            "solar_zenith_horn1",
                            "solar_zenith_horn2",
                            "solar_zenith_horn3",
                            "solar_zenith_horn4",
                            "satellite_azimuth_horn1",
                            "satellite_azimuth_horn2",
                            "satellite_azimuth_horn3",
                            "satellite_azimuth_horn4",
                            "surface_type_horn1",
                            "surface_type_horn2",
                            "surface_type_horn3",
                            "surface_type_horn4",
                            "terrain_elevation_horn1",
                            "terrain_elevation_horn2",
                            "terrain_elevation_horn3",
                            "terrain_elevation_horn4",
                            "longitude",
                            "latitude"]

class AWS_EPS_Sterna_BaseFileHandler(NetCDF4FileHandler):
    """Base class implementing the AWS/EPS-Sterna MWR Level-1b&c Filehandlers."""

    def __init__(self, filename, filename_info, filetype_info, auto_maskandscale=True):
        """Initialize the handler."""
        super().__init__(filename, filename_info, filetype_info,
                         cache_var_size=10000,
                         cache_handle=True)
        self.filename_info = filename_info

    @property
    def start_time(self):
        """Get the start time."""
        return self.filename_info["start_time"]

    @property
    def end_time(self):
        """Get the end time."""
        return self.filename_info["end_time"]

    @property
    def sensor(self):
        """Get the sensor name."""
        # This should have been self["/attr/instrument"]
        # But the sensor name is currently incorrect in the ESA level-1b files
        return "mwr"

    @property
    def platform_name(self):
        """Get the platform name."""
        return self.filename_info["platform_name"]

    @property
    def orbit_start(self):
        """Get the orbit number for the start of data."""
        return int(self["/attr/orbit_start"])

    @property
    def orbit_end(self):
        """Get the orbit number for the end of data."""
        return int(self["/attr/orbit_end"])

    def get_dataset(self, dataset_id, dataset_info):
        """Get the data."""
        raise NotImplementedError("This is not implemented in the Base class.")

    def _get_channel_data(self, dataset_id, dataset_info):
        channel_data = self[dataset_info["file_key"]]
        channel_data.coords["n_channels"] = MWR_CHANNEL_NAMES
        channel_data = channel_data.rename({"n_fovs": "x", "n_scans": "y"})
        return channel_data.sel(n_channels=dataset_id["name"]).drop_vars("n_channels")



class AWS_EPS_Sterna_MWR_L1BFile(AWS_EPS_Sterna_BaseFileHandler):
    """Class implementing the AWS/EPS-Sterna MWR L1b Filehandler."""

    def __init__(self, filename, filename_info, filetype_info, auto_maskandscale=True):
        """Initialize the handler."""
        super().__init__(filename, filename_info, filetype_info, auto_maskandscale)
        self._feed_horn_group_name = filetype_info.get("feed_horn_group_name")

    @property
    def sub_satellite_longitude_start(self):
        """Get the longitude of sub-satellite point at start of the product."""
        return self["status/satellite/subsat_longitude_start"].data.item()

    @property
    def sub_satellite_latitude_start(self):
        """Get the latitude of sub-satellite point at start of the product."""
        return self["status/satellite/subsat_latitude_start"].data.item()

    @property
    def sub_satellite_longitude_end(self):
        """Get the longitude of sub-satellite point at end of the product."""
        return self["status/satellite/subsat_longitude_end"].data.item()

    @property
    def sub_satellite_latitude_end(self):
        """Get the latitude of sub-satellite point at end of the product."""
        return self["status/satellite/subsat_latitude_end"].data.item()

    def get_dataset(self, dataset_id, dataset_info):
        """Get the data."""
        if dataset_id["name"] in MWR_CHANNEL_NAMES:
            data_array = self._get_channel_data(dataset_id, dataset_info)
        elif dataset_id["name"] in NAVIGATION_DATASET_NAMES:
            data_array = self._get_navigation_data(dataset_id, dataset_info)
        else:
            raise NotImplementedError(f"Dataset {dataset_id['name']} not available or not supported yet!")

        data_array = mask_and_scale(data_array)
        if dataset_id["name"] == "longitude":
            data_array = data_array.where(data_array <= 180, data_array - 360)

        data_array.attrs.update(dataset_info)

        data_array.attrs["orbital_parameters"] = {"sub_satellite_latitude_start": self.sub_satellite_latitude_start,
                                                  "sub_satellite_longitude_start": self.sub_satellite_longitude_start,
                                                  "sub_satellite_latitude_end": self.sub_satellite_latitude_end,
                                                  "sub_satellite_longitude_end": self.sub_satellite_longitude_end}

        data_array.attrs["platform_name"] = self.platform_name
        data_array.attrs["sensor"] = self.sensor
        data_array.attrs["orbit_number"] = self.orbit_start
        return data_array

    def _get_navigation_data(self, dataset_id, dataset_info):
        """Get the navigation (geolocation) data for one feed horn."""
        geo_data = self[dataset_info["file_key"]]
        geo_data.coords[self._feed_horn_group_name] = ["1", "2", "3", "4"]
        geo_data = geo_data.rename({"n_fovs": "x", "n_scans": "y"})
        horn = dataset_id["horn"].name
        _selection = {self._feed_horn_group_name: horn}
        return geo_data.sel(_selection).drop_vars(self._feed_horn_group_name)


def mask_and_scale(data_array):
    """Mask then scale the data array."""
    if "missing_value" in data_array.attrs:
        with xr.set_options(keep_attrs=True):
            data_array = data_array.where(data_array != data_array.attrs["missing_value"])
        data_array.attrs.pop("missing_value")
    if "valid_max" in data_array.attrs:
        with xr.set_options(keep_attrs=True):
            data_array = data_array.where(data_array <= data_array.attrs["valid_max"])
        data_array.attrs.pop("valid_max")
    if "valid_min" in data_array.attrs:
        with xr.set_options(keep_attrs=True):
            data_array = data_array.where(data_array >= data_array.attrs["valid_min"])
        data_array.attrs.pop("valid_min")
    if "scale_factor" in data_array.attrs and "add_offset" in data_array.attrs:
        with xr.set_options(keep_attrs=True):
            data_array = data_array * data_array.attrs["scale_factor"] + data_array.attrs["add_offset"]
        data_array.attrs.pop("scale_factor")
        data_array.attrs.pop("add_offset")
    return data_array



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2019 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Modis level 2 hdf-eos format reader.

Introduction
------------

The ``modis_l2`` reader reads and calibrates Modis L2 image data in hdf-eos format.
Since there are a multitude of different level 2 datasets not all of theses are implemented (yet).


Currently the reader supports:
    - m[o/y]d35_l2: cloud_mask dataset
    - some datasets in m[o/y]d06 files

Additionally the reader tries to add non yaml configured 2D datasets dynamically. As mentioned above there are a lot
of different level 2 datasets so this might not work in every case (for example bit encoded datasets similar to the
supported m[0/y]d35_l2 cloud mask are not decoded).

To get a list of the available datasets for a given file refer to the :ref:`reading:available datasets` section.


Geolocation files
-----------------

Similar to the ``modis_l1b`` reader the geolocation files (mod03) for the 1km data are optional and if not
given 1km geolocations will be interpolated from the 5km geolocation contained within the file.

For the 500m and 250m data geolocation files are needed.


References:
    - Documentation about the format: https://modis-atmos.gsfc.nasa.gov/products

"""
import logging

import dask.array as da
import numpy as np
import xarray as xr

from satpy.readers.core.hdf4 import from_sds
from satpy.readers.core.hdfeos import HDFEOSGeoReader
from satpy.utils import get_legacy_chunk_size

logger = logging.getLogger(__name__)
CHUNK_SIZE = get_legacy_chunk_size()


class ModisL2HDFFileHandler(HDFEOSGeoReader):
    """File handler for MODIS HDF-EOS Level 2 files.

    Includes error handling for files produced by IMAPP produced files.

    """

    def _load_all_metadata_attributes(self):
        try:
            return super()._load_all_metadata_attributes()
        except KeyError:
            return {}

    @property
    def is_imapp_mask_byte1(self):
        """Get if this file is the IMAPP 'mask_byte1' file type."""
        return "mask_byte1" in self.filetype_info["file_type"]

    @property
    def start_time(self):
        """Get the start time of the dataset."""
        try:
            return super().start_time
        except KeyError:
            try:
                return self.filename_info["start_time"]
            except KeyError:
                return self.filename_info["acquisition_time"]

    @property
    def end_time(self):
        """Get the end time of the dataset."""
        try:
            return super().end_time
        except KeyError:
            return self.start_time

    @staticmethod
    def read_geo_resolution(metadata):
        """Parse metadata to find the geolocation resolution.

        It is implemented as a staticmethod to match read_mda pattern.

        """
        try:
            return HDFEOSGeoReader.read_geo_resolution(metadata)
        except RuntimeError:
            # most L2 products are 5000m
            return 5000

    def _select_hdf_dataset(self, hdf_dataset_name, byte_dimension):
        """Load a dataset from HDF-EOS level 2 file."""
        dataset = self.sd.select(hdf_dataset_name)
        dask_arr = from_sds(dataset, self.filename, chunks=CHUNK_SIZE)
        attrs = dataset.attributes()
        dims = ["y", "x"]
        if byte_dimension == 0:
            dims = ["i", "y", "x"]
            dask_arr = dask_arr.astype(np.uint8)
        elif byte_dimension == 2:
            dims = ["y", "x", "i"]
            dask_arr = dask_arr.astype(np.uint8)
        dataset = xr.DataArray(dask_arr, dims=dims, attrs=attrs)
        if "i" in dataset.dims:
            # Reorder dimensions for consistency
            dataset = dataset.transpose("i", "y", "x")
        return dataset

    def get_dataset(self, dataset_id, dataset_info):
        """Get DataArray for specified dataset."""
        dataset_name = dataset_id["name"]
        if self.is_geo_loadable_dataset(dataset_name):
            return HDFEOSGeoReader.get_dataset(self, dataset_id, dataset_info)
        dataset_name_in_file = dataset_info["file_key"]
        if self.is_imapp_mask_byte1:
            dataset_name_in_file = dataset_info.get("imapp_file_key", dataset_name_in_file)

        # The dataset asked correspond to a given set of bits of the HDF EOS dataset
        if "byte" in dataset_info and "byte_dimension" in dataset_info:
            dataset = self._extract_and_mask_category_dataset(dataset_id, dataset_info, dataset_name_in_file)
        else:
            # No byte manipulation required
            dataset = self.load_dataset(dataset_name_in_file, dataset_info.pop("category", False))

        self._add_satpy_metadata(dataset_id, dataset)
        return dataset

    def available_datasets(self, configured_datasets):
        """Add dataset information from arbitrary level 2 files.

        Adds dataset information not specifically specified in reader yaml file
        from arbitrary modis level 2 product files to available datasets.

        Notes:
             Currently only adds 2D datasets and does not decode bit encoded information.
        """
        # pass along yaml configured (handled) datasets and collect their file keys to check against dynamically
        # collected variables later on.
        handled = set()
        for is_avail, ds_info in (configured_datasets or []):
            file_key = ds_info.get("file_key", ds_info["name"])
            handled.add(file_key)

            if is_avail is not None:
                yield is_avail, ds_info
                continue
            yield self.file_type_matches(ds_info["file_type"]), ds_info

        res_dict = {5416: 250, 2708: 500, 1354: 1000, 270: 5000, 135: 10000}

        # get variables from file dynamically and only add those which are not already configured in yaml
        for var_name, val in self.sd.datasets().items():
            if var_name in handled:
                continue
            if len(val[0]) != 2:
                continue
            resolution = res_dict.get(val[1][-1])
            if resolution is not None:
                ds_info = {
                    "file_type": self.filetype_info["file_type"],
                    "resolution": resolution,
                    "name": var_name,
                    "file_key": var_name,
                    "coordinates": ["longitude", "latitude"]
                }
                yield True, ds_info

    def _extract_and_mask_category_dataset(self, dataset_id, dataset_info, var_name):
        # what dimension is per-byte
        byte_dimension = None if self.is_imapp_mask_byte1 else dataset_info["byte_dimension"]
        dataset = self._select_hdf_dataset(var_name, byte_dimension)
        # category products always have factor=1/offset=0 so don't apply them
        # also remove them so they don't screw up future satpy processing
        dataset.attrs.pop("scale_factor", None)
        dataset.attrs.pop("add_offset", None)
        # Don't do this byte work if we are using the IMAPP mask_byte1 file
        if self.is_imapp_mask_byte1:
            return dataset

        dataset = _extract_byte_mask(dataset,
                                     dataset_info["byte"],
                                     dataset_info["bit_start"],
                                     dataset_info["bit_count"])
        dataset = self._mask_with_quality_assurance_if_needed(dataset, dataset_info, dataset_id)
        return dataset

    def _mask_with_quality_assurance_if_needed(self, dataset, dataset_info, dataset_id):
        if not dataset_info.get("quality_assurance", False):
            return dataset

        # Get quality assurance dataset recursively
        quality_assurance_dataset_id = dataset_id.from_dict(
            dict(name="quality_assurance", resolution=1000)
        )
        quality_assurance_dataset_info = {
            "name": "quality_assurance",
            "resolution": 1000,
            "byte_dimension": 2,
            "byte": 0,
            "bit_start": 0,
            "bit_count": 1,
            "file_key": "Quality_Assurance"
        }
        quality_assurance = self.get_dataset(
            quality_assurance_dataset_id, quality_assurance_dataset_info
        )
        # Duplicate quality assurance dataset to create relevant filter
        duplication_factor = [int(dataset_dim / quality_assurance_dim)
                              for dataset_dim, quality_assurance_dim
                              in zip(dataset.shape, quality_assurance.shape)]
        quality_assurance = np.tile(quality_assurance, duplication_factor)
        # Replace unassured data by NaN value
        dataset = dataset.where(quality_assurance != 0, dataset.attrs["_FillValue"])
        return dataset


def _extract_byte_mask(dataset, byte_information, bit_start, bit_count):
    attrs = dataset.attrs.copy()

    if isinstance(byte_information, int):
        # Only one byte: select the byte information
        byte_dataset = dataset[byte_information, :, :]
        dataset = _bits_strip(bit_start, bit_count, byte_dataset)
    elif isinstance(byte_information, (list, tuple)) and len(byte_information) == 2:
        # Two bytes: recombine the two bytes
        byte_mask = da.map_blocks(
            _extract_two_byte_mask,
            dataset.data[byte_information[0]],
            dataset.data[byte_information[1]],
            bit_start=bit_start,
            bit_count=bit_count,
            dtype=np.uint16,
            meta=np.array((), dtype=np.uint16),
            chunks=tuple(tuple(chunk_size * 4 for chunk_size in dim_chunks) for dim_chunks in dataset.chunks[1:]),
        )
        dataset = xr.DataArray(byte_mask, dims=dataset.dims[1:])

    # Compute the final bit mask
    dataset.attrs = attrs
    return dataset


def _extract_two_byte_mask(data_a: np.ndarray, data_b: np.ndarray, bit_start: int, bit_count: int) -> np.ndarray:
    data_a = data_a.astype(np.uint16, copy=False)
    data_a = np.left_shift(data_a, 8)  # dataset_a << 8
    byte_dataset = np.bitwise_or(data_a, data_b).astype(np.uint16)
    shape = byte_dataset.shape
    # We replicate the concatenated byte with the right shape
    byte_dataset = np.repeat(np.repeat(byte_dataset, 4, axis=0), 4, axis=1)
    # All bits carry information, we update bit_start consequently
    bit_start = np.arange(16, dtype=np.uint16).reshape((4, 4))
    bit_start = np.tile(bit_start, (shape[0], shape[1]))
    return _bits_strip(bit_start, bit_count, byte_dataset)


def _bits_strip(bit_start: int, bit_count: int, value: int) -> int:
    """Extract specified bit from bit representation of integer value.

    Args:
        bit_start:
            Starting index of the bits to extract (first bit has index 0)
        bit_count:
            Number of bits starting from bit_start to extract
        value:
            Number from which to extract the bits

    Returns:
        Value of the extracted bits

    """
    bit_mask = pow(2, bit_start + bit_count) - 1
    return np.right_shift(np.bitwise_and(value, bit_mask), bit_start)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2020-2021 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Classes for loading compositor and modifier configuration files."""
from __future__ import annotations

import logging
import os
import warnings
from functools import lru_cache, update_wrapper
from typing import Callable, Iterable

import yaml
from yaml import UnsafeLoader

import satpy
from satpy import DataID, DataQuery
from satpy._config import config_search_paths, get_entry_points_config_dirs, glob_config
from satpy.dataset.dataid import minimal_default_keys_config
from satpy.utils import recursive_dict_update

logger = logging.getLogger(__name__)


def _convert_dep_info_to_data_query(dep_info):
    key_item = dep_info.copy()
    key_item.pop("prerequisites", None)
    key_item.pop("optional_prerequisites", None)
    if "modifiers" in key_item:
        key_item["modifiers"] = tuple(key_item["modifiers"])
    key = DataQuery.from_dict(key_item)
    return key


class _CompositeConfigHelper:
    """Helper class for parsing composite configurations.

    The provided `loaded_compositors` dictionary is updated inplace.

    """

    def __init__(self, loaded_compositors, sensor_id_keys):
        self.loaded_compositors = loaded_compositors
        self.sensor_id_keys = sensor_id_keys

    def _create_comp_from_info(self, composite_info, loader):
        key = DataID(self.sensor_id_keys, **composite_info)
        comp = loader(_satpy_id=key, **composite_info)
        return key, comp

    def _handle_inline_comp_dep(self, dep_info, dep_num, parent_name):
        # Create an unique temporary name for the composite
        sub_comp_name = "_" + parent_name + "_dep_{}".format(dep_num)
        dep_info["name"] = sub_comp_name
        self._load_config_composite(dep_info)

    @staticmethod
    def _get_compositor_loader_from_config(composite_name, composite_info):
        try:
            loader = composite_info.pop("compositor")
        except KeyError:
            raise ValueError("'compositor' key missing or empty for '{}'. Option keys = {}".format(
                composite_name, str(composite_info.keys())))
        return loader

    def _process_composite_deps(self, composite_info):
        dep_num = -1
        for prereq_type in ["prerequisites", "optional_prerequisites"]:
            prereqs = []
            for dep_info in composite_info.get(prereq_type, []):
                dep_num += 1
                if not isinstance(dep_info, dict):
                    prereqs.append(dep_info)
                    continue
                elif "compositor" in dep_info:
                    self._handle_inline_comp_dep(
                        dep_info, dep_num, composite_info["name"])
                prereq_key = _convert_dep_info_to_data_query(dep_info)
                prereqs.append(prereq_key)
            composite_info[prereq_type] = prereqs

    def _load_config_composite(self, composite_info):
        composite_name = composite_info["name"]
        loader = self._get_compositor_loader_from_config(composite_name, composite_info)
        self._process_composite_deps(composite_info)
        key, comp = self._create_comp_from_info(composite_info, loader)
        self.loaded_compositors[key] = comp

    def _load_config_composites(self, configured_composites):
        for composite_name, composite_info in configured_composites.items():
            composite_info["name"] = composite_name
            self._load_config_composite(composite_info)

    def parse_config(self, configured_composites, composite_configs):
        """Parse composite configuration dictionary."""
        try:
            self._load_config_composites(configured_composites)
        except (ValueError, KeyError):
            raise RuntimeError("Failed to load composites from configs "
                               "'{}'".format(composite_configs))


class _ModifierConfigHelper:
    """Helper class for parsing modifier configurations.

    The provided `loaded_modifiers` dictionary is updated inplace.

    """

    def __init__(self, loaded_modifiers, sensor_id_keys):
        self.loaded_modifiers = loaded_modifiers
        self.sensor_id_keys = sensor_id_keys

    @staticmethod
    def _get_modifier_loader_from_config(modifier_name, modifier_info):
        try:
            loader = modifier_info.pop("modifier", None)
            if loader is None:
                loader = modifier_info.pop("compositor")
                warnings.warn(
                    "Modifier '{}' uses deprecated 'compositor' "
                    "key to point to Python class, replace "
                    "with 'modifier'.".format(modifier_name),
                    stacklevel=5
                )
        except KeyError:
            raise ValueError("'modifier' key missing or empty for '{}'. Option keys = {}".format(
                modifier_name, str(modifier_info.keys())))
        return loader

    def _process_modifier_deps(self, modifier_info):
        for prereq_type in ["prerequisites", "optional_prerequisites"]:
            prereqs = []
            for dep_info in modifier_info.get(prereq_type, []):
                if not isinstance(dep_info, dict):
                    prereqs.append(dep_info)
                    continue
                prereq_key = _convert_dep_info_to_data_query(dep_info)
                prereqs.append(prereq_key)
            modifier_info[prereq_type] = prereqs

    def _load_config_modifier(self, modifier_info):
        modifier_name = modifier_info["name"]
        loader = self._get_modifier_loader_from_config(modifier_name, modifier_info)
        self._process_modifier_deps(modifier_info)
        self.loaded_modifiers[modifier_name] = (loader, modifier_info)

    def _load_config_modifiers(self, configured_modifiers):
        for modifier_name, modifier_info in configured_modifiers.items():
            modifier_info["name"] = modifier_name
            self._load_config_modifier(modifier_info)

    def parse_config(self, configured_modifiers, composite_configs):
        """Parse modifier configuration dictionary."""
        try:
            self._load_config_modifiers(configured_modifiers)
        except (ValueError, KeyError):
            raise RuntimeError("Failed to load modifiers from configs "
                               "'{}'".format(composite_configs))


def _load_config(composite_configs):
    if not isinstance(composite_configs, (list, tuple)):
        composite_configs = [composite_configs]

    conf = {}
    for composite_config in composite_configs:
        with open(composite_config, "r", encoding="utf-8") as conf_file:
            recursive_dict_update(conf, yaml.load(conf_file, Loader=UnsafeLoader))
    try:
        sensor_name = conf["sensor_name"]
    except KeyError:
        logger.debug('No "sensor_name" tag found in %s, skipping.',
                     composite_configs)
        return {}, {}, {}

    sensor_compositors = {}
    sensor_modifiers = {}

    dep_id_keys = None
    sensor_deps = sensor_name.split("/")[:-1]
    if sensor_deps:
        # get dependent
        for sensor_dep in sensor_deps:
            dep_comps, dep_mods, dep_id_keys = load_compositor_configs_for_sensor(sensor_dep)
        # the last parent should include all of its parents so only add the last one
        sensor_compositors.update(dep_comps)
        sensor_modifiers.update(dep_mods)

    id_keys = _get_sensor_id_keys(conf, dep_id_keys)
    mod_config_helper = _ModifierConfigHelper(sensor_modifiers, id_keys)
    configured_modifiers = conf.get("modifiers", {})
    mod_config_helper.parse_config(configured_modifiers, composite_configs)

    comp_config_helper = _CompositeConfigHelper(sensor_compositors, id_keys)
    configured_composites = conf.get("composites", {})
    comp_config_helper.parse_config(configured_composites, composite_configs)
    return sensor_compositors, sensor_modifiers, id_keys


def _get_sensor_id_keys(conf, parent_id_keys):
    try:
        id_keys = conf["composite_identification_keys"]
    except KeyError:
        id_keys = parent_id_keys
        if not id_keys:
            id_keys = minimal_default_keys_config
    return id_keys


def _lru_cache_with_config_path(func: Callable):
    """Use lru_cache but include satpy's current config_path."""
    @lru_cache()
    def _call_without_config_path_wrapper(sensor_name, _):
        return func(sensor_name)

    def _add_config_path_wrapper(sensor_name: str):
        config_path = satpy.config.get("config_path")
        # make sure config_path is hashable, but keep original order since it matters
        config_path = tuple(config_path)
        return _call_without_config_path_wrapper(sensor_name, config_path)

    wrapper = update_wrapper(_add_config_path_wrapper, func)
    wrapper = _update_cached_wrapper(wrapper, _call_without_config_path_wrapper)
    return wrapper


def _update_cached_wrapper(wrapper, cached_func):
    for meth_name in ("cache_clear", "cache_parameters", "cache_info"):
        if hasattr(cached_func, meth_name):
            setattr(wrapper, meth_name, getattr(cached_func, meth_name))
    return wrapper


@_lru_cache_with_config_path
def load_compositor_configs_for_sensor(sensor_name: str) -> tuple[dict[str, dict], dict[str, dict], dict]:
    """Load compositor, modifier, and DataID key information from configuration files for the specified sensor.

    Args:
        sensor_name: Sensor name that has matching ``sensor_name.yaml``
            config files.

    Returns:
        tuple: (comps, mods, data_id_keys), Where `comps` is a dictionary:

                composite ID -> compositor object

            And `mods` is a dictionary:

                modifier name -> (modifier class, modifiers options)

            Add `data_id_keys` is a dictionary:

                DataID key -> key properties

    """
    config_filename = sensor_name + ".yaml"
    logger.debug("Looking for composites config file %s", config_filename)
    paths = get_entry_points_config_dirs("satpy.composites")
    composite_configs = config_search_paths(
        os.path.join("composites", config_filename),
        search_dirs=paths, check_exists=True)
    if not composite_configs:
        logger.debug("No composite config found called %s",
                     config_filename)
        return {}, {}, minimal_default_keys_config
    return _load_config(composite_configs)


def load_compositor_configs_for_sensors(sensor_names: Iterable[str]) -> tuple[dict[str, dict], dict[str, dict]]:
    """Load compositor and modifier configuration files for the specified sensors.

    Args:
        sensor_names: Sensor names that have matching
            ``sensor_name.yaml`` config files.

    Returns:
        tuple: (comps, mods), Where `comps` is a dictionary:

                sensor_name -> composite ID -> compositor object

            And `mods` is a dictionary:

                sensor_name -> modifier name -> (modifier class,
                modifiers options)

    """
    comps = {}
    mods = {}
    for sensor_name in sensor_names:
        sensor_comps, sensor_mods = load_compositor_configs_for_sensor(sensor_name)[:2]
        comps[sensor_name] = sensor_comps
        mods[sensor_name] = sensor_mods
    return comps, mods


def all_composite_sensors():
    """Get all sensor names from available composite configs."""
    paths = get_entry_points_config_dirs("satpy.composites")
    composite_configs = glob_config(
        os.path.join("composites", "*.yaml"),
        search_dirs=paths)
    yaml_names = set([os.path.splitext(os.path.basename(fn))[0]
                      for fn in composite_configs])
    non_sensor_yamls = ("visir",)
    sensor_names = [x for x in yaml_names if x not in non_sensor_yamls]
    return sensor_names



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2020 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Classes and functions related to a dictionary with DataID keys."""

import numpy as np

from .dataid import DataID, create_filtered_query, minimal_default_keys_config


class TooManyResults(KeyError):
    """Special exception when one key maps to multiple items in the container."""


def get_best_dataset_key(key, choices):
    """Choose the "best" `DataID` from `choices` based on `key`.

    To see how the keys are sorted, refer to `:meth:satpy.datasets.DataQuery.sort_dataids`.

    This function assumes `choices` has already been filtered to only
    include datasets that match the provided `key`.

    Args:
        key (DataQuery): Query parameters to sort `choices` by.
        choices (Iterable): `DataID` objects to sort through to determine
                            the best dataset.

    Returns: List of best `DataID`s from `choices`. If there is more
             than one element this function could not choose between the
             available datasets.

    """
    sorted_choices, distances = key.sort_dataids(choices)
    if len(sorted_choices) == 0 or distances[0] is np.inf:
        return []
    else:
        return [choice for choice, distance in zip(sorted_choices, distances) if distance == distances[0]]


def get_key(key, key_container, num_results=1, best=True, query=None,  # noqa: D417
            **kwargs):
    """Get the fully-specified key best matching the provided key.

    Only the best match is returned if `best` is `True` (default). See
    `get_best_dataset_key` for more information on how this is determined.

    `query` is provided as a convenience to filter by multiple parameters
    at once without having to filter by multiple `key` inputs.

    Args:
        key (DataID): DataID of query parameters to use for
                         searching. Any parameter that is `None`
                         is considered a wild card and any match is
                         accepted.
        key_container (dict or set): Container of DataID objects that
                                     uses hashing to quickly access items.
        num_results (int): Number of results to return. Use `0` for all
                           matching results. If `1` then the single matching
                           key is returned instead of a list of length 1.
                           (default: 1)
        best (bool): Sort results to get "best" result first
                     (default: True). See `get_best_dataset_key` for details.
        query (DataQuery): filter for the key which can contain for example:

            resolution (float, int, or list): Resolution of the dataset in
                                            dataset units (typically
                                            meters). This can also be a
                                            list of these numbers.
            calibration (str or list): Dataset calibration
                                    (ex.'reflectance'). This can also be a
                                    list of these strings.
            polarization (str or list): Dataset polarization
                                        (ex.'V'). This can also be a
                                        list of these strings.
            level (number or list): Dataset level (ex. 100). This can also be a
                                    list of these numbers.
            modifiers (list): Modifiers applied to the dataset. Unlike
                            resolution and calibration this is the exact
                            desired list of modifiers for one dataset, not
                            a list of possible modifiers.

    Returns:
        list or DataID: Matching key(s)

    Raises: KeyError if no matching results or if more than one result is
            found when `num_results` is `1`.

    """
    key = create_filtered_query(key, query)

    res = key.filter_dataids(key_container)
    if not res:
        raise KeyError("No dataset matching '{}' found".format(str(key)))

    if best:
        res = get_best_dataset_key(key, res)

    if num_results == 1 and not res:
        raise KeyError("No dataset matching '{}' found".format(str(key)))
    if num_results == 1 and len(res) != 1:
        raise TooManyResults("No unique dataset matching {}".format(str(key)))
    if num_results == 1:
        return res[0]
    if num_results == 0:
        return res

    return res[:num_results]


class DatasetDict(dict):
    """Special dictionary object that can handle dict operations based on dataset name, wavelength, or DataID.

    Note: Internal dictionary keys are `DataID` objects.

    """

    def keys(self, names=False, wavelengths=False):
        """Give currently contained keys."""
        # sort keys so things are a little more deterministic (.keys() is not)
        keys = sorted(super(DatasetDict, self).keys())
        if names:
            return (k.get("name") for k in keys)
        elif wavelengths:
            return (k.get("wavelength") for k in keys)
        else:
            return keys

    def get_key(self, match_key, num_results=1, best=True, **dfilter):  # noqa: D417
        """Get multiple fully-specified keys that match the provided query.

        Args:
            key (DataID): DataID of query parameters to use for
                          searching. Any parameter that is `None`
                          is considered a wild card and any match is
                          accepted. Can also be a string representing the
                          dataset name or a number representing the dataset
                          wavelength.
            num_results (int): Number of results to return. If `0` return all,
                               if `1` return only that element, otherwise
                               return a list of matching keys.
            **dfilter (dict): See `get_key` function for more information.

        """
        return get_key(match_key, self.keys(), num_results=num_results,
                       best=best, **dfilter)

    def getitem(self, item):
        """Get Node when we know the *exact* DataID."""
        return super(DatasetDict, self).__getitem__(item)

    def __getitem__(self, item):
        """Get item from container."""
        try:
            # short circuit - try to get the object without more work
            return super(DatasetDict, self).__getitem__(item)
        except KeyError:
            key = self.get_key(item)
            return super(DatasetDict, self).__getitem__(key)

    def get(self, key, default=None):
        """Get value with optional default."""
        try:
            key = self.get_key(key)
        except KeyError:
            return default
        return super(DatasetDict, self).get(key, default)

    def __setitem__(self, key, value):
        """Support assigning 'Dataset' objects or dictionaries of metadata."""
        if hasattr(value, "attrs"):
            # xarray.DataArray objects
            value_info = value.attrs
        else:
            value_info = value
        # use value information to make a more complete DataID
        if not isinstance(key, DataID):
            key = self._create_dataid_key(key, value_info)

        # update the 'value' with the information contained in the key
        try:
            new_info = key.to_dict()
        except AttributeError:
            new_info = key
        if isinstance(value_info, dict):
            value_info.update(new_info)
            if isinstance(key, DataID):
                value_info["_satpy_id"] = key

        return super(DatasetDict, self).__setitem__(key, value)

    def _create_dataid_key(self, key, value_info):
        """Create a DataID key from dictionary."""
        if not isinstance(value_info, dict):
            raise ValueError("Key must be a DataID when value is not an xarray DataArray or dict")
        old_key = key
        try:
            key = self.get_key(key)
        except KeyError:
            if isinstance(old_key, str):
                new_name = old_key
            else:
                new_name = value_info.get("name")
            # this is a new key and it's not a full DataID tuple
            if new_name is None and value_info.get("wavelength") is None:
                raise ValueError("One of 'name' or 'wavelength' attrs "
                                 "values should be set.")
            id_keys = self._create_id_keys_from_dict(value_info)
            value_info["name"] = new_name
            key = DataID(id_keys, **value_info)
        return key

    def _create_id_keys_from_dict(self, value_info_dict):
        """Create id_keys from dict."""
        try:
            id_keys = value_info_dict["_satpy_id"].id_keys
        except KeyError:
            try:
                id_keys = value_info_dict["_satpy_id_keys"]
            except KeyError:
                id_keys = minimal_default_keys_config
        return id_keys

    def contains(self, item):
        """Check contains when we know the *exact* DataID."""
        return super(DatasetDict, self).__contains__(item)

    def __contains__(self, item):
        """Check if item exists in container."""
        try:
            key = self.get_key(item)
        except KeyError:
            return False
        return super(DatasetDict, self).__contains__(key)

    def __delitem__(self, key):
        """Delete item from container."""
        try:
            # short circuit - try to get the object without more work
            return super(DatasetDict, self).__delitem__(key)
        except KeyError:
            key = self.get_key(key)
            return super(DatasetDict, self).__delitem__(key)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2019 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""Geostationary Lightning Mapper reader for the Level 2 format from glmtools.

More information about `glmtools` and the files it produces can be found on
the project's GitHub repository:

    https://github.com/deeplycloudy/glmtools

"""

import datetime as dt
import logging

import numpy as np

from satpy.readers.core.abi import NC_ABI_BASE

logger = logging.getLogger(__name__)

PLATFORM_NAMES = {
    "G16": "GOES-16",
    "G17": "GOES-17",
    "G18": "GOES-18",
}

# class NC_GLM_L2_LCFA(BaseFileHandler): — add this with glmtools


class NCGriddedGLML2(NC_ABI_BASE):
    """File reader for individual GLM L2 NetCDF4 files."""

    @property
    def sensor(self):
        """Get sensor name for current file handler."""
        return "glm"

    @property
    def start_time(self):
        """Start time of the current file's observations."""
        return dt.datetime.strptime(self.nc.attrs["time_coverage_start"], "%Y-%m-%dT%H:%M:%SZ")

    @property
    def end_time(self):
        """End time of the current file's observations."""
        return dt.datetime.strptime(self.nc.attrs["time_coverage_end"], "%Y-%m-%dT%H:%M:%SZ")

    def _is_category_product(self, data_arr):
        # if after autoscaling we still have an integer
        is_int = np.issubdtype(data_arr.dtype, np.integer)
        # and it has a fill value
        has_fill = "_FillValue" in data_arr.attrs
        # or it has flag_meanings
        has_meanings = "flag_meanings" in data_arr.attrs
        # then it is likely a category product and we should keep the
        # _FillValue for satpy to use later
        return is_int and (has_fill or has_meanings)

    def get_dataset(self, key, info):
        """Load a dataset."""
        logger.debug("Reading in get_dataset %s.", key["name"])
        res = self[key["name"]]
        res.attrs.update({"platform_name": self.platform_name,
                          "sensor": self.sensor})
        res.attrs.update(self.filename_info)

        # Add orbital parameters
        projection = self.nc["goes_imager_projection"]
        res.attrs["orbital_parameters"] = {
            "projection_longitude": float(projection.attrs["longitude_of_projection_origin"]),
            "projection_latitude": float(projection.attrs["latitude_of_projection_origin"]),
            "projection_altitude": float(projection.attrs["perspective_point_height"]),
            "satellite_nominal_latitude": float(self["nominal_satellite_subpoint_lat"]),
            "satellite_nominal_longitude": float(self["nominal_satellite_subpoint_lon"]),
            # 'satellite_nominal_altitude': float(self['nominal_satellite_height']),
        }

        res.attrs.update(key.to_dict())

        # remove attributes that could be confusing later
        if not self._is_category_product(res):
            res.attrs.pop("_FillValue", None)
        res.attrs.pop("scale_factor", None)
        res.attrs.pop("add_offset", None)
        res.attrs.pop("_Unsigned", None)
        res.attrs.pop("ancillary_variables", None)  # Can't currently load DQF
        # add in information from the filename that may be useful to the user
        # for key in ('observation_type', 'scene_abbr', 'scan_mode', 'platform_shortname'):
        for attr in ("scene_abbr", "scan_mode", "platform_shortname"):
            res.attrs[attr] = self.filename_info[attr]
        # copy global attributes to metadata
        for attr in ("scene_id", "orbital_slot", "instrument_ID",
                     "production_site", "timeline_ID", "spatial_resolution"):
            res.attrs[attr] = self.nc.attrs.get(attr)
        return res

    def _is_2d_xy_var(self, data_arr):
        is_2d = data_arr.ndim == 2
        has_x_dim = "x" in data_arr.dims
        has_y_dim = "y" in data_arr.dims
        return is_2d and has_x_dim and has_y_dim

    def available_datasets(self, configured_datasets=None):
        """Discover new datasets and add information from file."""
        # we know the actual resolution
        res = self.spatial_resolution_to_number()

        # update previously configured datasets
        handled_vars = set()
        for is_avail, ds_info in (configured_datasets or []):
            # some other file handler knows how to load this
            # don't override what they've done
            if is_avail is not None:
                yield is_avail, ds_info

            matches = self.file_type_matches(ds_info["file_type"])
            if matches and ds_info.get("resolution") != res:
                # we are meant to handle this dataset (file type matches)
                # and the information we can provide isn't available yet
                new_info = ds_info.copy()
                new_info["resolution"] = res
                exists = ds_info["name"] in self.nc
                handled_vars.add(ds_info["name"])
                yield exists, new_info
            elif is_avail is None:
                # we don't know what to do with this
                # see if another future file handler does
                yield is_avail, ds_info

        for var_name, data_arr in self.nc.data_vars.items():
            if var_name in handled_vars:
                # it was manually configured and handled above
                continue
            if not self._is_2d_xy_var(data_arr):
                # only handle 2d (y, x) vars for now
                continue

            new_info = {
                "name": var_name,
                "resolution": res,
                "file_type": self.filetype_info["file_type"]
            }
            handled_vars.add(var_name)
            yield True, new_info



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2016-2018 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Test for readers/virr_l1b.py."""
import os
import unittest
from unittest import mock

import dask.array as da
import numpy as np
import xarray as xr

from satpy.tests.reader_tests.test_hdf5_utils import FakeHDF5FileHandler


class FakeHDF5FileHandler2(FakeHDF5FileHandler):
    """Swap-in HDF5 File Handler."""

    def make_test_data(self, dims):
        """Create fake test data."""
        return xr.DataArray(da.from_array(np.ones([dim for dim in dims], dtype=np.float32) * 10, [dim for dim in dims]))

    def _make_file(self, platform_id, geolocation_prefix, l1b_prefix, ECWN, Emissive_units):
        dim_0 = 19
        dim_1 = 20
        test_file = {
            # Satellite data.
            "/attr/Day Or Night Flag": "D", "/attr/Observing Beginning Date": "2018-12-25",
            "/attr/Observing Beginning Time": "21:41:47.090", "/attr/Observing Ending Date": "2018-12-25",
            "/attr/Observing Ending Time": "21:47:28.254", "/attr/Satellite Name": platform_id,
            "/attr/Sensor Identification Code": "VIRR",
            # Emissive data.
            l1b_prefix + "EV_Emissive": self.make_test_data([3, dim_0, dim_1]),
            l1b_prefix + "EV_Emissive/attr/valid_range": [0, 50000],
            l1b_prefix + "Emissive_Radiance_Scales": self.make_test_data([dim_0, dim_1]),
            l1b_prefix + "EV_Emissive/attr/units": Emissive_units,
            l1b_prefix + "Emissive_Radiance_Offsets": self.make_test_data([dim_0, dim_1]),
            "/attr/" + ECWN: [2610.31, 917.6268, 836.2546],
            # Reflectance data.
            l1b_prefix + "EV_RefSB": self.make_test_data([7, dim_0, dim_1]),
            l1b_prefix + "EV_RefSB/attr/valid_range": [0, 32767], l1b_prefix + "EV_RefSB/attr/units": "none",
            "/attr/RefSB_Cal_Coefficients": np.ones(14, dtype=np.float32) * 2
        }
        for attribute in ["Latitude", "Longitude", geolocation_prefix + "SolarZenith",
                          geolocation_prefix + "SensorZenith", geolocation_prefix + "SolarAzimuth",
                          geolocation_prefix + "SensorAzimuth"]:
            test_file[attribute] = self.make_test_data([dim_0, dim_1])
            test_file[attribute + "/attr/Intercept"] = 0.
            test_file[attribute + "/attr/units"] = "degrees"
            if "Solar" in attribute or "Sensor" in attribute:
                test_file[attribute + "/attr/Slope"] = .01
                if "Azimuth" in attribute:
                    test_file[attribute + "/attr/valid_range"] = [0, 18000]
                else:
                    test_file[attribute + "/attr/valid_range"] = [-18000, 18000]
            else:
                test_file[attribute + "/attr/Slope"] = 1.
                if "Longitude" == attribute:
                    test_file[attribute + "/attr/valid_range"] = [-180., 180.]
                else:
                    test_file[attribute + "/attr/valid_range"] = [-90., 90.]
        return test_file

    def get_test_content(self, filename, filename_info, filetype_info):
        """Mimic reader input file content."""
        if filename_info["platform_id"] == "FY3B":
            return self._make_file("FY3B", "", "", "Emmisive_Centroid_Wave_Number", "milliWstts/m^2/cm^(-1)/steradian")
        return self._make_file(filename_info["platform_id"], "Geolocation/", "Data/",
                               "Emissive_Centroid_Wave_Number", "none")


class TestVIRRL1BReader(unittest.TestCase):
    """Test VIRR L1B Reader."""

    yaml_file = "virr_l1b.yaml"

    def setUp(self):
        """Wrap HDF5 file handler with our own fake handler."""
        from satpy._config import config_search_paths
        from satpy.readers.virr_l1b import VIRR_L1B
        self.reader_configs = config_search_paths(os.path.join("readers", self.yaml_file))
        # http://stackoverflow.com/questions/12219967/how-to-mock-a-base-class-with-python-mock-library
        self.p = mock.patch.object(VIRR_L1B, "__bases__", (FakeHDF5FileHandler2,))
        self.fake_handler = self.p.start()
        self.p.is_local = True

    def tearDown(self):
        """Stop wrapping the HDF5 file handler."""
        self.p.stop()

    def _band_helper(self, attributes, units, calibration, standard_name,
                     file_type, band_index_size, resolution):
        assert units == attributes["units"]
        assert calibration == attributes["calibration"]
        assert standard_name == attributes["standard_name"]
        assert file_type == attributes["file_type"]
        assert (attributes["band_index"] in range(band_index_size)) is True
        assert resolution == attributes["resolution"]
        assert ("longitude", "latitude") == attributes["coordinates"]

    def _fy3_helper(self, platform_name, reader, Emissive_units):
        """Load channels and test accurate metadata."""
        import datetime
        band_values = {"1": 22.0, "2": 22.0, "6": 22.0, "7": 22.0, "8": 22.0, "9": 22.0, "10": 22.0,
                       "3": 496.542155, "4": 297.444511, "5": 288.956557, "solar_zenith_angle": .1,
                       "satellite_zenith_angle": .1, "solar_azimuth_angle": .1, "satellite_azimuth_angle": .1,
                       "longitude": 10}
        if platform_name == "FY3B":
            # updated 2015 coefficients
            band_values["1"] = -0.168
            band_values["2"] = -0.2706
            band_values["6"] = -1.5631
            band_values["7"] = -0.2114
            band_values["8"] = -0.171
            band_values["9"] = -0.1606
            band_values["10"] = -0.1328
        datasets = reader.load([band for band in band_values])
        for dataset in datasets:
            # Object returned by get_dataset.
            ds = datasets[dataset["name"]]
            attributes = ds.attrs
            assert isinstance(ds.data, da.Array)
            assert "virr" == attributes["sensor"]
            assert platform_name == attributes["platform_name"]
            assert datetime.datetime(2018, 12, 25, 21, 41, 47, 90000) == attributes["start_time"]
            assert datetime.datetime(2018, 12, 25, 21, 47, 28, 254000) == attributes["end_time"]
            assert (19, 20) == datasets[dataset["name"]].shape
            assert ("y", "x") == datasets[dataset["name"]].dims
            if dataset["name"] in ["1", "2", "6", "7", "8", "9", "10"]:
                self._band_helper(attributes, "%", "reflectance",
                                  "toa_bidirectional_reflectance", "virr_l1b",
                                  7, 1000)
            elif dataset["name"] in ["3", "4", "5"]:
                self._band_helper(attributes, Emissive_units, "brightness_temperature",
                                  "toa_brightness_temperature", "virr_l1b", 3, 1000)
            elif dataset["name"] in ["longitude", "latitude"]:
                assert "degrees" == attributes["units"]
                assert (attributes["standard_name"] in ["longitude", "latitude"]) is True
                assert ["virr_l1b", "virr_geoxx"] == attributes["file_type"]
                assert 1000 == attributes["resolution"]
            else:
                assert "degrees" == attributes["units"]
                assert attributes["standard_name"] in ["solar_zenith_angle", "sensor_zenith_angle",
                                                       "solar_azimuth_angle", "sensor_azimuth_angle"]
                assert ["virr_geoxx", "virr_l1b"] == attributes["file_type"]
                assert ("longitude", "latitude") == attributes["coordinates"]
            np.testing.assert_allclose(band_values[dataset["name"]], ds[ds.shape[0] // 2][ds.shape[1] // 2], rtol=1e-6)
            assert "valid_range" not in ds.attrs

    def test_fy3b_file(self):
        """Test that FY3B files are recognized."""
        from satpy.readers.core.loading import load_reader
        FY3B_reader = load_reader(self.reader_configs)
        FY3B_file = FY3B_reader.select_files_from_pathnames(["tf2018359214943.FY3B-L_VIRRX_L1B.HDF"])
        assert 1 == len(FY3B_file)
        FY3B_reader.create_filehandlers(FY3B_file)
        # Make sure we have some files
        assert FY3B_reader.file_handlers
        self._fy3_helper("FY3B", FY3B_reader, "milliWstts/m^2/cm^(-1)/steradian")

    def test_fy3c_file(self):
        """Test that FY3C files are recognized."""
        from satpy.readers.core.loading import load_reader
        FY3C_reader = load_reader(self.reader_configs)
        FY3C_files = FY3C_reader.select_files_from_pathnames(["tf2018359143912.FY3C-L_VIRRX_GEOXX.HDF",
                                                              "tf2018359143912.FY3C-L_VIRRX_L1B.HDF"])
        assert 2 == len(FY3C_files)
        FY3C_reader.create_filehandlers(FY3C_files)
        # Make sure we have some files
        assert FY3C_reader.file_handlers
        self._fy3_helper("FY3C", FY3C_reader, "1")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2016-2021 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Interface to NUCAPS Retrieval NetCDF files.

NUCAPS stands for NOAA Unique Combined Atmospheric Processing System.
NUCAPS retrievals include temperature, moisture, trace gas, and cloud-cleared
radiance profiles. Product details can be found at:

https://www.ospo.noaa.gov/Products/atmosphere/soundings/nucaps/

This reader supports both standard NOAA NUCAPS EDRs, and Science EDRs,
which are essentially a subset of the standard EDRs with some additional
parameters such as relative humidity and boundary layer temperature.

NUCAPS data is derived from Cross-track Infrared Sounder (CrIS) data, and
from Advanced Technology Microwave Sounder (ATMS) data, instruments
onboard Joint Polar Satellite System spacecraft.

"""

import logging
from collections import defaultdict

import numpy as np
import pandas as pd
import xarray as xr

from satpy.readers.core.netcdf import NetCDF4FileHandler
from satpy.readers.core.yaml_reader import FileYAMLReader

LOG = logging.getLogger(__name__)

# It's difficult to do processing without knowing the pressure levels beforehand
ALL_PRESSURE_LEVELS = [
    0.0161, 0.0384, 0.0769, 0.137, 0.2244, 0.3454, 0.5064, 0.714,
    0.9753, 1.2972, 1.6872, 2.1526, 2.7009, 3.3398, 4.077, 4.9204,
    5.8776, 6.9567, 8.1655, 9.5119, 11.0038, 12.6492, 14.4559, 16.4318,
    18.5847, 20.9224, 23.4526, 26.1829, 29.121, 32.2744, 35.6505,
    39.2566, 43.1001, 47.1882, 51.5278, 56.126, 60.9895, 66.1253,
    71.5398, 77.2396, 83.231, 89.5204, 96.1138, 103.017, 110.237,
    117.777, 125.646, 133.846, 142.385, 151.266, 160.496, 170.078,
    180.018, 190.32, 200.989, 212.028, 223.441, 235.234, 247.408,
    259.969, 272.919, 286.262, 300, 314.137, 328.675, 343.618, 358.966,
    374.724, 390.893, 407.474, 424.47, 441.882, 459.712, 477.961,
    496.63, 515.72, 535.232, 555.167, 575.525, 596.306, 617.511, 639.14,
    661.192, 683.667, 706.565, 729.886, 753.628, 777.79, 802.371,
    827.371, 852.788, 878.62, 904.866, 931.524, 958.591, 986.067,
    1013.95, 1042.23, 1070.92, 1100
]


class NUCAPSFileHandler(NetCDF4FileHandler):
    """File handler for NUCAPS netCDF4 format."""

    def __init__(self, *args, **kwargs):
        """Initialize file handler."""
        # remove kwargs that reader instance used that file handler does not
        kwargs.pop("mask_surface", None)
        kwargs.pop("mask_quality", None)
        kwargs.setdefault("xarray_kwargs", {}).setdefault(
            "decode_times", False)
        super(NUCAPSFileHandler, self).__init__(*args, **kwargs)

    def __contains__(self, item):
        """Return item from file content."""
        return item in self.file_content

    def _parse_datetime(self, datestr):
        """Parse NUCAPS datetime string."""
        return pd.to_datetime(datestr).to_pydatetime()

    @property
    def start_time(self):
        """Get start time."""
        try:
            return self._parse_datetime(self["/attr/time_coverage_start"])
        except KeyError:
            # If attribute not present, use time from file name
            return self.filename_info["start_time"]

    @property
    def end_time(self):
        """Get end time."""
        try:
            return self._parse_datetime(self["/attr/time_coverage_end"])
        except KeyError:
            # If attribute not present, use time from file name
            return self.filename_info["end_time"]

    @property
    def start_orbit_number(self):
        """Return orbit number for the beginning of the swath."""
        try:
            return int(self["/attr/start_orbit_number"])
        except KeyError:
            return 0

    @property
    def end_orbit_number(self):
        """Return orbit number for the end of the swath."""
        try:
            return int(self["/attr/end_orbit_number"])
        except KeyError:
            return 0

    @property
    def platform_name(self):
        """Return standard platform name for the file's data."""
        try:
            res = self["/attr/platform_name"]
            if isinstance(res, np.ndarray):
                return str(res.astype(str))
            return res
        except KeyError:
            return self.filename_info["platform_shortname"]

    @property
    def sensor_names(self):
        """Return standard sensor or instrument name for the file's data."""
        try:
            res = self["/attr/instrument_name"]
            res = [x.strip() for x in res.split(",")]
            if len(res) == 1:
                return res[0].lower()
        except KeyError:
            res = ["CrIS", "ATMS", "VIIRS"]
        return set(name.lower() for name in res)

    def get_shape(self, ds_id, ds_info):
        """Return data array shape for item specified."""
        var_path = ds_info.get("file_key", "{}".format(ds_id["name"]))
        if var_path + "/shape" not in self:
            # loading a scalar value
            shape = 1
        else:
            shape = self[var_path + "/shape"]
            if "index" in ds_info:
                shape = shape[1:]
            if "pressure_index" in ds_info:
                shape = shape[:-1]
        return shape

    def get_metadata(self, dataset_id, ds_info):
        """Get metadata."""
        var_path = ds_info.get("file_key", "{}".format(dataset_id["name"]))
        shape = self.get_shape(dataset_id, ds_info)
        file_units = ds_info.get("file_units",
                                 self.get(var_path + "/attr/units"))
        ds_info.update(getattr(self[var_path], "attrs", {}))
        # don't overwrite information in the files attrs because the same
        # `.attrs` is used for each separate Temperature pressure level dataset
        # Plus, if someone gets metadata multiple times then we are screwed
        info = ds_info
        info.update(ds_info)
        info.update(dataset_id.to_dict())
        info.update({
            "shape": shape,
            "units": ds_info.get("units", file_units),
            "platform_name": self.platform_name,
            "sensor": self.sensor_names,
            "start_orbit": self.start_orbit_number,
            "end_orbit": self.end_orbit_number,
        })
        if "standard_name" not in info:
            sname_path = var_path + "/attr/standard_name"
            info["standard_name"] = self.get(sname_path)
        if dataset_id["name"] != "Quality_Flag":
            anc_vars = info.get("ancillary_variables", [])
            if "Quality_Flag" not in anc_vars:
                anc_vars.append("Quality_Flag")
                info["ancillary_variables"] = anc_vars
        return info

    def get_dataset(self, dataset_id, ds_info):
        """Load data array and metadata for specified dataset."""
        var_path = ds_info.get("file_key", "{}".format(dataset_id["name"]))
        metadata = self.get_metadata(dataset_id, ds_info)
        valid_min, valid_max = self[var_path + "/attr/valid_range"]
        fill_value = self.get(var_path + "/attr/_FillValue")

        d_tmp = self[var_path]
        if "index" in ds_info:
            d_tmp = d_tmp[int(ds_info["index"])]
        if "pressure_index" in ds_info:
            d_tmp = d_tmp[..., int(ds_info["pressure_index"])]
            # this is a pressure based field
            # include surface_pressure as metadata
            sp = self["Surface_Pressure"]
            # Older format
            if "number_of_FORs" in sp.dims:
                sp = sp.rename({"number_of_FORs": "y"})
            # Newer format
            if "Number_of_CrIS_FORs" in sp.dims:
                sp = sp.rename({"Number_of_CrIS_FORs": "y"})
            if "surface_pressure" in ds_info:
                ds_info["surface_pressure"] = xr.concat((ds_info["surface_pressure"], sp), dim="y")
            else:
                ds_info["surface_pressure"] = sp
            # include all the pressure levels
            ds_info.setdefault("pressure_levels", self["Pressure"][0])
        data = d_tmp

        if valid_min is not None and valid_max is not None:
            # the original .cfg/INI based reader only checked valid_max
            data = data.where((data <= valid_max))  # | (data >= valid_min))
        if fill_value is not None:
            data = data.where(data != fill_value)
            # this _FillValue is no longer valid
            metadata.pop("_FillValue", None)
            data.attrs.pop("_FillValue", None)

        data.attrs.update(metadata)
        # Older format
        if "number_of_FORs" in data.dims:
            data = data.rename({"number_of_FORs": "y"})
        # Newer format
        if "Number_of_CrIS_FORs" in data.dims:
            data = data.rename({"Number_of_CrIS_FORs": "y"})
        return data


class NUCAPSReader(FileYAMLReader):
    """Reader for NUCAPS NetCDF4 files."""

    def __init__(self, config_files, mask_surface=True, mask_quality=True, **kwargs):  # noqa: D417
        """Configure reader behavior.

        Args:
            mask_surface (bool): mask anything below the surface pressure
            mask_quality (bool): mask anything where the `Quality_Flag` metadata is ``!= 1``.

        """
        self.pressure_dataset_names = defaultdict(list)
        super(NUCAPSReader, self).__init__(config_files,
                                           **kwargs)
        self.mask_surface = self.info.get("mask_surface", mask_surface)
        self.mask_quality = self.info.get("mask_quality", mask_quality)

    def load_ds_ids_from_config(self):
        """Convert config dataset entries to DataIDs.

        Special handling is done to provide level specific datasets
        for any pressured based datasets. For example, a dataset is
        added for each pressure level of 'Temperature' with each
        new dataset being named 'Temperature_Xmb' where X is the
        pressure level.
        """
        super(NUCAPSReader, self).load_ds_ids_from_config()
        for ds_id in list(self.all_ids.keys()):
            ds_info = self.all_ids[ds_id]
            if ds_info.get("pressure_based", False):
                for idx, lvl_num in enumerate(ALL_PRESSURE_LEVELS):
                    if lvl_num < 5.0:
                        suffix = "_{:0.03f}mb".format(lvl_num)
                    else:
                        suffix = "_{:0.0f}mb".format(lvl_num)

                    new_info = ds_info.copy()
                    new_info["pressure_level"] = lvl_num
                    new_info["pressure_index"] = idx
                    new_info["file_key"] = "{}".format(ds_id["name"])
                    new_info["name"] = ds_id["name"] + suffix
                    new_ds_id = ds_id._replace(name=new_info["name"])
                    new_info["id"] = new_ds_id
                    self.all_ids[new_ds_id] = new_info
                    self.pressure_dataset_names[ds_id["name"]].append(new_info["name"])

    def load(self, dataset_keys, previous_datasets=None, pressure_levels=None):
        """Load data from one or more set of files.

        :param pressure_levels: mask out certain pressure levels:
                                True for all levels
                                (min, max) for a range of pressure levels
                                [...] list of levels to include
        """
        dataset_keys = set(self.get_dataset_key(x) for x in dataset_keys)
        if pressure_levels is not None:
            self._filter_dataset_keys_outside_pressure_levels(dataset_keys, pressure_levels)

            # Add pressure levels to the datasets to load if needed so
            # we can do further filtering after loading
            plevels_ds_id = self.get_dataset_key("Pressure_Levels")
            remove_plevels = False
            if plevels_ds_id not in dataset_keys:
                dataset_keys.add(plevels_ds_id)
                remove_plevels = True

        datasets_loaded = super(NUCAPSReader, self).load(
            dataset_keys, previous_datasets=previous_datasets)

        if pressure_levels is not None:
            if remove_plevels:
                plevels_ds = datasets_loaded.pop(plevels_ds_id)
                dataset_keys.remove(plevels_ds_id)
            else:
                plevels_ds = datasets_loaded[plevels_ds_id]
            _remove_data_at_pressure_levels(datasets_loaded, plevels_ds, pressure_levels)

        if self.mask_surface:
            _mask_data_below_surface_pressure(datasets_loaded, dataset_keys)

        if self.mask_quality:
            _mask_data_with_quality_flag(datasets_loaded, dataset_keys)

        return datasets_loaded

    def _filter_dataset_keys_outside_pressure_levels(self, dataset_keys, pressure_levels):
        for ds_id in dataset_keys.copy():
            ds_info = self.all_ids[ds_id]
            ds_level = ds_info.get("pressure_level")
            if ds_level is not None:
                if pressure_levels is True:
                    # they want all pressure levels
                    continue
                elif len(pressure_levels) == 2 and pressure_levels[0] <= ds_level <= pressure_levels[1]:
                    # given a min and a max pressure level
                    continue
                elif np.isclose(pressure_levels, ds_level).any():
                    # they asked for this specific pressure level
                    continue
                else:
                    # they don't want this dataset at this pressure level
                    LOG.debug("Removing dataset to load: %s", ds_id)
                    dataset_keys.remove(ds_id)
                    continue


def _remove_data_at_pressure_levels(datasets_loaded, plevels_ds, pressure_levels):
    cond = _get_pressure_level_condition(plevels_ds, pressure_levels)
    if cond is not None:
        new_plevels = plevels_ds.where(cond, drop=True)
    else:
        new_plevels = plevels_ds
    for ds_id in datasets_loaded.keys():
        ds_obj = datasets_loaded[ds_id]
        if plevels_ds.dims[0] not in ds_obj.dims:
            continue

        if cond is not None:
            datasets_loaded[ds_id] = ds_obj.where(cond, drop=True)
        datasets_loaded[ds_id].attrs["pressure_levels"] = new_plevels


def _get_pressure_level_condition(plevels_ds, pressure_levels):
    if pressure_levels is True:
        return None
    if len(pressure_levels) == 2:
        cond = (plevels_ds >= pressure_levels[0]) & (plevels_ds <= pressure_levels[1])
    else:
        cond = plevels_ds == pressure_levels
    # convert dask-based DataArray to a computed numpy-based DataArray to
    # avoid unknown shapes of dask arrays when this condition is used for masking
    return cond.compute()


def _mask_data_below_surface_pressure(datasets_loaded, dataset_keys):
    LOG.debug("Filtering pressure levels at or below the surface pressure")
    for ds_id in sorted(dataset_keys):
        ds = datasets_loaded[ds_id]
        if "surface_pressure" not in ds.attrs or "pressure_levels" not in ds.attrs:
            continue
        data_pressure = ds.attrs["pressure_levels"]
        surface_pressure = ds.attrs["surface_pressure"]
        if isinstance(surface_pressure, float):
            # scalar needs to become array for each record
            surface_pressure = np.repeat(surface_pressure, ds.shape[0])
        if surface_pressure.ndim == 1 and surface_pressure.shape[0] == ds.shape[0]:
            # surface is one element per record
            LOG.debug("Filtering %s at and below the surface pressure", ds_id)
            if ds.ndim == 2:
                surface_pressure = np.repeat(surface_pressure[:, None], data_pressure.shape[0], axis=1)
                data_pressure = np.repeat(data_pressure[None, :], surface_pressure.shape[0], axis=0)
                datasets_loaded[ds_id] = ds.where(data_pressure < surface_pressure)
            else:
                # entire dataset represents one pressure level
                data_pressure = ds.attrs["pressure_level"]
                datasets_loaded[ds_id] = ds.where(data_pressure < surface_pressure)
        else:
            LOG.warning("Not sure how to handle shape of 'surface_pressure' metadata")


def _mask_data_with_quality_flag(datasets_loaded, dataset_keys):
    LOG.debug("Filtering data based on quality flags")
    for ds_id in sorted(dataset_keys):
        ds = datasets_loaded[ds_id]
        quality_flag = [
            x for x in ds.attrs.get("ancillary_variables", [])
            if x.attrs.get("name") == "Quality_Flag"]
        if not quality_flag:
            continue

        quality_flag = quality_flag[0]
        if quality_flag.dims[0] not in ds.dims:
            continue
        LOG.debug("Masking %s where quality flag doesn't equal 1", ds_id)
        datasets_loaded[ds_id] = ds.where(quality_flag == 0)



#!/usr/bin/python
# Copyright (c) 2016 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Unittests for coordinate utilities."""

import unittest

import dask.array as da
import numpy as np
import xarray as xr
from pyproj import CRS


class TestCoordinateHelpers(unittest.TestCase):
    """Test various utility functions for working with coordinates."""

    def test_area_def_coordinates(self):
        """Test coordinates being added with an AreaDefinition."""
        from pyresample.geometry import AreaDefinition

        from satpy.coords import add_crs_xy_coords
        area_def = AreaDefinition(
            "test", "test", "test", {"proj": "lcc", "lat_1": 25, "lat_0": 25},
            100, 200, [-100, -100, 100, 100]
        )
        data_arr = xr.DataArray(
            da.zeros((200, 100), chunks=50),
            attrs={"area": area_def},
            dims=("y", "x"),
        )
        new_data_arr = add_crs_xy_coords(data_arr, area_def)
        assert "y" in new_data_arr.coords
        assert "x" in new_data_arr.coords

        assert "units" in new_data_arr.coords["y"].attrs
        assert new_data_arr.coords["y"].attrs["units"] == "meter"
        assert "units" in new_data_arr.coords["x"].attrs
        assert new_data_arr.coords["x"].attrs["units"] == "meter"
        assert "crs" in new_data_arr.coords
        assert isinstance(new_data_arr.coords["crs"].item(), CRS)
        assert area_def.crs == new_data_arr.coords["crs"].item()

        # already has coords
        data_arr = xr.DataArray(
            da.zeros((200, 100), chunks=50),
            attrs={"area": area_def},
            dims=("y", "x"),
            coords={"y": np.arange(2, 202), "x": np.arange(100)}
        )
        new_data_arr = add_crs_xy_coords(data_arr, area_def)
        assert "y" in new_data_arr.coords
        assert "units" not in new_data_arr.coords["y"].attrs
        assert "x" in new_data_arr.coords
        assert "units" not in new_data_arr.coords["x"].attrs
        np.testing.assert_equal(new_data_arr.coords["y"], np.arange(2, 202))

        assert "crs" in new_data_arr.coords
        assert isinstance(new_data_arr.coords["crs"].item(), CRS)
        assert area_def.crs == new_data_arr.coords["crs"].item()

        # lat/lon area
        area_def = AreaDefinition(
            "test", "test", "test", {"proj": "latlong"},
            100, 200, [-100, -100, 100, 100]
        )
        data_arr = xr.DataArray(
            da.zeros((200, 100), chunks=50),
            attrs={"area": area_def},
            dims=("y", "x"),
        )
        new_data_arr = add_crs_xy_coords(data_arr, area_def)
        assert "y" in new_data_arr.coords
        assert "x" in new_data_arr.coords

        assert "units" in new_data_arr.coords["y"].attrs
        assert new_data_arr.coords["y"].attrs["units"] == "degrees_north"
        assert "units" in new_data_arr.coords["x"].attrs
        assert new_data_arr.coords["x"].attrs["units"] == "degrees_east"
        assert "crs" in new_data_arr.coords
        assert isinstance(new_data_arr.coords["crs"].item(), CRS)
        assert area_def.crs == new_data_arr.coords["crs"].item()

    def test_swath_def_coordinates(self):
        """Test coordinates being added with an SwathDefinition."""
        from pyresample.geometry import SwathDefinition

        from satpy.coords import add_crs_xy_coords
        lons_data = da.random.random((200, 100), chunks=50)
        lats_data = da.random.random((200, 100), chunks=50)
        lons = xr.DataArray(lons_data, attrs={"units": "degrees_east"},
                            dims=("y", "x"))
        lats = xr.DataArray(lats_data, attrs={"units": "degrees_north"},
                            dims=("y", "x"))
        area_def = SwathDefinition(lons, lats)
        data_arr = xr.DataArray(
            da.zeros((200, 100), chunks=50),
            attrs={"area": area_def},
            dims=("y", "x"),
        )
        new_data_arr = add_crs_xy_coords(data_arr, area_def)
        # See https://github.com/pydata/xarray/issues/3068
        # self.assertIn('longitude', new_data_arr.coords)
        # self.assertIn('units', new_data_arr.coords['longitude'].attrs)
        # self.assertEqual(
        #     new_data_arr.coords['longitude'].attrs['units'], 'degrees_east')
        # self.assertIsInstance(new_data_arr.coords['longitude'].data, da.Array)
        # self.assertIn('latitude', new_data_arr.coords)
        # self.assertIn('units', new_data_arr.coords['latitude'].attrs)
        # self.assertEqual(
        #     new_data_arr.coords['latitude'].attrs['units'], 'degrees_north')
        # self.assertIsInstance(new_data_arr.coords['latitude'].data, da.Array)

        assert "crs" in new_data_arr.coords
        crs = new_data_arr.coords["crs"].item()
        assert isinstance(crs, CRS)
        assert crs.is_geographic
        assert isinstance(new_data_arr.coords["crs"].item(), CRS)
